@@ -23,7 +23,7 @@ from typing import Optional
 
 import click
 
-from . import DefaultNames, __version__, get_home_config_path
+from . import DefaultNames, __version__, get_home_config_path, is_version_unreleased
 from . import init as initmod
 from . import logs, version_tuple
 from .job import start_job
@@ -64,7 +64,10 @@ def option_group(*options):
     is_flag=True,
     help="Only output errors to the stdout",
 )
-@click.option("--logfile", default=None, help="Log messages to file (at DEBUG level)")
+@click.option("--logfile",
+  default=None,
+  envvar="UNFURL_LOGFILE",
+  help="Log messages to file (at DEBUG level)")
 @click.option(
     "--tmp",
     envvar="UNFURL_TMPDIR",
@@ -75,7 +78,12 @@ def option_group(*options):
 @click.option(
     "--version-check",
     envvar="UNFURL_VERSION_CHECK",
-    help="warn if older than the given version",
+    help="Abort if the runtime's Unfurl is older than the given version",
+)
+@click.option(
+    "--no-version-check",
+    is_flag=True,
+    help="Skip the Unfurl version check when invoking the runtime.",
 )
 def cli(
     ctx,
@@ -84,28 +92,36 @@ def cli(
     loglevel=None,
     tmp=None,
     version_check=None,
+    home=None,
     **kw,
 ):
     # ensure that ctx.obj exists and is a dict (in case `cli()` is called
     # by means other than the `if` block below
     ctx.ensure_object(dict)
+    if not kw.get("logfile"):
+        kw["logfile"] = logs.get_tmplog_path()
     ctx.obj.update(kw)
-
     if tmp is not None:
         os.environ["UNFURL_TMPDIR"] = tmp
+    if home is not None:
+        os.environ["UNFURL_HOME"] = home
     effective_log_level = detect_log_level(loglevel, quiet, verbose)
     ctx.obj["verbose"] = detect_verbose_level(effective_log_level)
-    if kw["logfile"]:
-        logs.add_log_file(kw["logfile"])
+    logs.add_log_file(kw["logfile"], effective_log_level)
     logs.set_console_log_level(effective_log_level)
 
     if version_check and version_tuple() < version_tuple(version_check):
-        logging.warning(
-            "current version %s older than expected version %s",
+        logging.error(
+            "Current unfurl version %s older than expected version %s",
             __version__(True),
             version_check,
         )
-
+        if is_version_unreleased(version_check):
+            msg = f"{sys.executable} -m pip install -U 'git+https://github.com/onecommons/unfurl.git#egg=unfurl'"
+        else:
+            msg = f"{sys.executable} -m pip install -U unfurl=={version_check}"
+        logging.error("Use --no-version-check to ignore or run this command to upgrade: \n %s", msg)
+        raise click.Abort()
 
 def detect_log_level(loglevel: Optional[str], quiet: bool, verbose: int) -> Levels:
     if quiet:
@@ -159,7 +175,12 @@ readonlyJobControlOptions = option_group(
         "--jobexitcode",
         type=click.Choice(["error", "degraded", "never"]),
         default="never",
-        help="Set exit code to 1 if job status is not ok.",
+        help="Set exit code to 1 if job status is not ok. (Default: never)",
+    ),
+    click.option(
+        "--use-environment",
+        default=None,
+        help="Run this job in the given environment.",
     ),
 )
 jobControlOptions = option_group(
@@ -261,16 +282,17 @@ def _venv(runtime, env):
         env = os.environ.copy()
     # see virtualenv activate
     env.pop("PYTHONHOME", None)  # unset if set
+    runtime = os.path.expanduser(runtime)
     env["VIRTUAL_ENV"] = runtime
     env["PATH"] = os.path.join(runtime, "bin") + os.pathsep + env.get("PATH", "")
     return env
 
 
-def _remote_cmd(runtime_, cmd_line, local_env):
+def _remote_cmd(runtime_, cmd_line, local_env, version_check):
     context = local_env.get_context()
     kind, sep, rest = runtime_.partition(":")
     envvar_filter = context.get("variables") or {}
-    for name in ["UNFURL_APPROVE", "UNFURL_LOGGING"]:
+    for name in ["UNFURL_APPROVE", "UNFURL_LOGGING", "UNFURL_MOCK_DEPLOY"]:
         if name in os.environ:
             envvar_filter[name] = os.environ[name]
 
@@ -289,21 +311,20 @@ def _remote_cmd(runtime_, cmd_line, local_env):
                 "-m",
                 "unfurl",
                 "--no-runtime",
-                "--version-check",
-                __version__(True),
             ]
+            + version_check
             + cmd_line,
             False,
         )
     elif kind == "docker":
         cmd = DockerCmd(runtime_, env or {}).build()
-        return env, cmd + cmd_line, False
+        return env, cmd + version_check + cmd_line, False
     else:
         # treat as shell command
         cmd = shlex.split(runtime_)
         return (
             env,
-            cmd + ["--no-runtime", "--version-check", __version__(True)] + cmd_line,
+            cmd + ["--no-runtime"] + version_check + cmd_line,
             True,
         )
 
@@ -355,7 +376,7 @@ class DockerCmd:
         cmd.extend(self.default_volumes())
         cmd.extend(self.docker_args)
         cmd.extend(
-            [self.image, "unfurl", "--no-runtime", "--version-check", __version__(True)]
+            [self.image, "unfurl", "--no-runtime"]
         )
         return cmd
 
@@ -387,11 +408,18 @@ class DockerCmd:
 
 def _run_remote(runtime, options, localEnv):
     logger = logging.getLogger("unfurl")
-    logger.debug('running command remotely on "%s"', runtime)
+    logger.info('running command remotely on "%s"', runtime)
     cmdLine = _args or sys.argv[1:]
     if _args:  # set by test driver to override command line
         print(f"TESTING: running remote with _args {_args}")
-    env, remote, shell = _remote_cmd(runtime, cmdLine, localEnv)
+
+    if options.get("no_version_check"):
+        cmdLine.remove('--no-version-check')
+        version_check = []
+    else:
+        version_check = ["--version-check", __version__(True)]
+
+    env, remote, shell = _remote_cmd(runtime, cmdLine, localEnv, version_check)
     logger.debug("executing remote command: %s", remote)
     rv = subprocess.call(remote, env=env, shell=shell)
     if options.get("standalone_mode") is False:
@@ -435,33 +463,27 @@ def yesno(prompt):
 
 
 def _stop_logging(job, options, verbose, tmplogfile):
-    if options["logfile"]:
-        if verbose > -1:
-            click.echo("Done, full log appended to " + options["logfile"])
+    if job and job.log_path:
+        log_path = job.log_path
+        dir = os.path.dirname(log_path)
+        if not os.path.exists(dir):
+            os.makedirs(dir)
+        try:
+            os.rename(tmplogfile, log_path)
+        except OSError:
+            # handle [Errno 18] Invalid cross-device link
+            shutil.copy(tmplogfile, log_path)
     else:
-        if job and job.log_path:
-            log_path = job.log_path
-            dir = os.path.dirname(log_path)
-            if not os.path.exists(dir):
-                os.makedirs(dir)
-            try:
-                os.rename(tmplogfile, log_path)
-            except OSError:
-                # handle [Errno 18] Invalid cross-device link
-                shutil.copy(tmplogfile, log_path)
-        else:
-            log_path = tmplogfile
-        if verbose > -1:
-            click.echo("Done, full log written to " + log_path)
+        log_path = tmplogfile
+    if verbose > -1:
+        click.echo("Done, full log written to " + log_path)
 
 
 def _run_local(ensemble, options):
+    logger = logging.getLogger("unfurl")
+    logger.verbose('Running command: %s', sys.argv[1:])
     verbose = options.get("verbose", 0)
-    tmplogfile = None
-    if not options["logfile"]:
-        tmplogfile = logs.get_tmplog_path()
-        logs.add_log_file(tmplogfile)
-
+    tmplogfile = options["logfile"]
     job, rendered, proceed = start_job(ensemble, options)
     _latestJobs.append(job)  # testing only
     if job:
@@ -684,6 +706,10 @@ def plan(ctx, ensemble=None, **options):
     default=False,
     help="Generate files only (don't commit them).",
 )
+@click.option(
+    "--use-deployment-blueprint",
+    help="Use this deployment blueprint.",
+)
 def init(ctx, projectdir, ensemble_name=None, **options):
     """
     Create a new project or, if [project_dir] exists or is inside a project, create a new ensemble.
@@ -721,6 +747,10 @@ def init(ctx, projectdir, ensemble_name=None, **options):
                 'Can not create project in "' + projectdir + '": folder is not empty'
             )
 
+    if options.get("use_deployment_blueprint"):
+        raise click.ClickException(
+            'Can not use "--use-deployment-blueprint" option when creating a new project.'
+        )
     homePath, projectPath, repo = initmod.create_project(
         os.path.abspath(projectdir), ensemble_name, **options
     )
@@ -842,6 +872,10 @@ def runtime(ctx, project_folder, init=False, **options):
     is_flag=True,
     help="Create ensemble in the given directory even if it exists.",
 )
+@click.option(
+    "--use-deployment-blueprint",
+    help="Use this deployment blueprint.",
+)
 def clone(ctx, source, dest, **options):
     """Create a new ensemble or project from a service template or an existing ensemble or project.
 
@@ -993,15 +1027,22 @@ def git_status(ctx, project_or_ensemble_path, dirty, **options):
     default="deployment",
     type=click.Choice(["blueprint", "environments", "deployment"]),
 )
+@click.option(
+    "--use-environment",
+    default=None,
+    help="Export using this environment.",
+)
 def export(ctx, project_or_ensemble_path, format, **options):
     """Export ensemble in a simplified json format."""
     from . import to_json
 
     options.update(ctx.obj)
-    localEnv = LocalEnv(project_or_ensemble_path, options.get("home"))
+    localEnv = LocalEnv(project_or_ensemble_path, options.get("home"), override_context=options.get("use_environment"))
     exporter = getattr(to_json, "to_" + format)
     jsonSummary = exporter(localEnv)
     click.echo(json.dumps(jsonSummary, indent=2))
+    logger = logging.getLogger("unfurl")
+    logger.info('Export complete.')
 
 
 @cli.command()
@@ -1010,7 +1051,7 @@ def export(ctx, project_or_ensemble_path, format, **options):
 @click.option("--query", help="Run the given expression")
 @click.option("--trace", default=0, help="Set the query's trace level")
 def status(ctx, ensemble, **options):
-    """"Show the status of deployed resources in the given ensemble."""
+    """ "Show the status of deployed resources in the given ensemble."""
     options.update(ctx.obj)
     localEnv = LocalEnv(ensemble, options.get("home"))
     manifest = localEnv.get_manifest()
