@@ -1,11 +1,22 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
+from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Union, ValuesView, cast
+from typing_extensions import TypedDict
 import six
 from collections.abc import Mapping, MutableSequence
 import os
 import copy
+
+from unfurl.logs import UnfurlLogger
+if TYPE_CHECKING:
+    from unfurl.manifest import Manifest
+    from unfurl.job import ConfigTask
+
+
 from .support import Status, ResourceChanges, Priority, TopologyMap
 from .result import (
+    ChangeRecord,
+    ResultsList,
     serialize_value,
     ChangeAware,
     Results,
@@ -22,10 +33,11 @@ from .util import (
     to_enum,
     wrap_sensitive_value,
     sensitive,
+    truncate_str
 )
 from . import merge
 from .eval import Ref, map_value, RefContext
-from .runtime import RelationshipInstance, Operational
+from .runtime import NodeInstance, RelationshipInstance, Operational
 from .yamlloader import yaml
 from .projectpaths import WorkFolder, Folders
 from .planrequests import (
@@ -39,8 +51,8 @@ from .planrequests import (
 
 import logging
 
-logger = logging.getLogger("unfurl.task")
-
+# Tell mypy the logger is of type UnfurlLogger
+logger = cast(UnfurlLogger, logging.getLogger("unfurl.task"))
 
 class ConfiguratorResult:
     """
@@ -52,14 +64,14 @@ class ConfiguratorResult:
 
     def __init__(
         self,
-        success,
-        modified,
-        status=None,
-        configChanged=None,
-        result=None,
-        outputs=None,
-        exception=None,
-    ):
+        success: bool,
+        modified: Optional[bool],
+        status: Optional[Status]=None,
+        configChanged: bool=None,
+        result: object=None,
+        outputs: Optional[dict]=None,
+        exception: Optional[UnfurlTaskError]=None,
+    ) -> None:
         self.modified = modified
         self.status = to_enum(Status, status)
         self.configChanged = configChanged
@@ -68,7 +80,7 @@ class ConfiguratorResult:
         self.outputs = outputs
         self.exception = None
 
-    def __str__(self):
+    def __str__(self) -> str:
         result = "" if self.result is None else str(self.result)[:240] + "..."
         return (
             "changes: "
@@ -77,9 +89,10 @@ class ConfiguratorResult:
                     filter(
                         None,
                         [
-                            self.success and "success",
-                            self.modified and "modified",
-                            self.status is not None and self.status.name,
+                            # mypy isn't happy with the following lines, but they should work fine
+                            self.success and "success",  # type: ignore
+                            self.modified and "modified",  # type: ignore
+                            self.status is not None and self.status.name,  # type: ignore
                         ],
                     )
                 )
@@ -91,10 +104,10 @@ class ConfiguratorResult:
 
 
 class AutoRegisterClass(type):
-    def __new__(mcls, name, bases, dct):
+    def __new__(mcls, name: str, bases: tuple, dct: dict) -> "type":
         cls = type.__new__(mcls, name, bases, dct)
-        if cls.short_name:
-            name = cls.short_name
+        if cls.short_name:  # type: ignore
+            name = cls.short_name  # type: ignore
         elif name.endswith("Configurator"):
             name = name[: -len("Configurator")]
         if name:
@@ -105,7 +118,7 @@ class AutoRegisterClass(type):
 @six.add_metaclass(AutoRegisterClass)
 class Configurator:
 
-    short_name = None
+    short_name: Optional[str] = None
     """shortName can be used to customize the "short name" of the configurator
     as an alternative to using the full name ("module.class") when setting the implementation on an operation.
     (Titlecase recommended)"""
@@ -113,16 +126,16 @@ class Configurator:
     exclude_from_digest = ()
 
     @classmethod
-    def set_config_spec_args(klass, kw: dict, target):
+    def set_config_spec_args(klass, kw: dict, target: None) -> dict:
         return kw
 
-    def __init__(self, configurationSpec):
+    def __init__(self, configurationSpec: ConfigurationSpec) -> None:
         self.configSpec = configurationSpec
 
-    def get_generator(self, task):
+    def get_generator(self, task: "TaskView") -> Generator:
         return self.run(task)
 
-    def render(self, task):
+    def render(self, task: "TaskView") -> None:
         """
         This method is called is called during the planning phase to give the configurator an
         opportunity to do early validation and error detection and generate any plan information or configuration files that the user may want to review before the running the deployment task.
@@ -135,7 +148,7 @@ class Configurator:
         return None
 
     # yields a JobRequest, TaskRequest or a ConfiguratorResult
-    def run(self, task):
+    def run(self, task: "TaskView") -> Generator:
         """
         This should perform the operation specified in the :class:`ConfigurationSpec`
         on the :obj:`task.target`.
@@ -149,7 +162,7 @@ class Configurator:
         """
         yield task.done(False)
 
-    def can_dry_run(self, task):
+    def can_dry_run(self, task: "TaskView") -> bool:
         """
         Returns whether this configurator can handle a dry-runs for the given task.
         (And should check :attr:`.TaskView.dry_run` in during run().
@@ -162,7 +175,7 @@ class Configurator:
         """
         return False
 
-    def can_run(self, task):
+    def can_run(self, task: "TaskView") -> Union[bool, str]:
         """
         Return whether or not the configurator can execute the given task
         depending on if this configurator support the requested action and parameters
@@ -176,11 +189,11 @@ class Configurator:
         """
         return True
 
-    def should_run(self, task):
+    def should_run(self, task: "TaskView") -> bool:
         """Does this configuration need to be run?"""
         return self.configSpec.should_run()
 
-    def save_digest(self, task):
+    def save_digest(self, task: "TaskView") -> dict:
         """
         Generate a compact, deterministic representation of the current configuration.
         This is saved in the job log and used by `check_digest` in subsequent jobs to
@@ -197,7 +210,8 @@ class Configurator:
             dict: A dictionary whose keys are strings that start with "digest"
         """
         # XXX user definition should be able to exclude inputs from digest
-        inputs = task._resolved_inputs
+        # XXX might throw AttributeError
+        inputs = task._resolved_inputs  # type: ignore
 
         # sensitive values are always redacted so no point in including them in the digest
         # (for cleaner output and security-in-depth)
@@ -225,7 +239,7 @@ class Configurator:
         )
         return digest
 
-    def check_digest(self, task, changeset):
+    def check_digest(self, task: "TaskView", changeset: ChangeRecord) -> bool:
         """
         Examine the previous :class:`ChangeRecord` generated by the previous time this operation
         was performed on the target instance and return whether it should be rerun or not.
@@ -246,10 +260,27 @@ class Configurator:
         newKeys = {k for k in task.inputs.keys() if k not in self.exclude_from_digest}
         task.logger.debug("checking digest for %s: %s", task.target.name, _parameters)
         if not _parameters:
-            return bool(newKeys)
-        keys = _parameters.split(",")
+            if newKeys:
+                task.logger.verbose(
+                    "digest keys for %s was previously empty, now found %s",
+                    task.target.name,
+                    newKeys,
+                )
+                return True
+            else:
+                return False
+        else:
+            keys = _parameters.split(",")
         oldInputs = {key for key in keys if "::" not in key}
+        # XXX this check is incomplete because this isn't considering new inputs because we don't know if they will be resolved or not
+        # we could fix this by having digest keys include unresolved inputs
         if oldInputs - newKeys:
+            task.logger.verbose(
+                "digest keys changed for %s: old %s, new %s",
+                task.target.name,
+                oldInputs,
+                newKeys,
+            )
             return True  # an old input was removed
 
         # only resolve the inputs and dependencies that were resolved before
@@ -261,20 +292,28 @@ class Configurator:
                 results.append(task.inputs._getresult(key))
 
         newDigest = get_digest(results, manifest=task._manifest)
-        mismatch = changeset.digestValue != newDigest
+        # I can't find somewhere where digestValue is defined on ChangeRecord, but I'll assume the use is correct
+        mismatch = changeset.digestValue != newDigest  # type: ignore
         if mismatch:
             task.logger.verbose(
                 "digests didn't match for %s with %s: old %s, new %s",
                 task.target.name,
                 _parameters,
-                changeset.digestValue,
+                changeset.digestValue, # type: ignore
                 newDigest,
             )
         return mismatch
 
 
+class MockConfigurator(Configurator):
+    def run(self, task: "TaskView") -> Generator:
+        yield task.done(True)
+
+    def can_dry_run(self, task: "TaskView") -> bool:
+        return True
+
 class _ConnectionsMap(dict):
-    def by_type(self):
+    def by_type(self) -> ValuesView:
         # return unique connection by type
         # reverse so nearest relationships replace less specific ones that have matching names
         by_type = {  # the list() is for Python 3.7
@@ -282,7 +321,7 @@ class _ConnectionsMap(dict):
         }
         return by_type.values()
 
-    def __missing__(self, key):
+    def __missing__(self, key: object) -> object:
         # the more specific connections are inserted first so this should find
         # the most relevant connection of the given type
         for value in self.values():
@@ -310,7 +349,13 @@ class TaskView:
         verbose (int): Verbosity level set for this job (-1 error, 0 normal, 1 verbose, 2 debug)
     """
 
-    def __init__(self, manifest, configSpec, target, reason=None, dependencies=None):
+    def __init__(self,
+        manifest: "Manifest",
+        configSpec: ConfigurationSpec,
+        target: NodeInstance,
+        reason: Optional[str]=None,
+        dependencies: Optional[List["Dependency"]]=None
+    ) -> None:
         # public:
         self.configSpec = configSpec
         self.target = target
@@ -318,22 +363,24 @@ class TaskView:
         self.logger = logger
         self.cwd = os.path.abspath(self.target.base_dir)
         self.rendered = None
+        self.dry_run = None
         # private:
-        self._errors = []  # UnfurlTaskError objects appends themselves to this list
-        self._inputs = None
+        self._errors: List[UnfurlTaskError] = []  # UnfurlTaskError objects appends themselves to this list
+        self._inputs: Optional[ResultsMap] = None
         self._environ = None
         self._manifest = manifest
-        self.messages = []
-        self._addedResources = []
+        self.messages: List[object] = []
+        self._addedResources: List[NodeInstance] = []
         self._dependenciesChanged = False
         self.dependencies = dependencies or []
         self._resourceChanges = ResourceChanges()
-        self._workFolders = {}
+        self._workFolders: dict = {}
+        self._rendering = False
         # public:
         self.operation_host = find_operation_host(target, configSpec.operation_host)
 
     @property
-    def inputs(self):
+    def inputs(self) -> ResultsMap:
         """
         Exposes inputs and task settings as expression variables, so they can be accessed like:
 
@@ -344,13 +391,13 @@ class TaskView:
         {{ inputs.param }}
         """
         if self._inputs is None:
-            assert self._attributeManager
-            assert self.target.root.attributeManager is self._attributeManager
+            assert self._attributeManager  # type: ignore
+            assert self.target.root.attributeManager is self._attributeManager  # type: ignore
             # deepcopy because ResultsMap might modify interior maps and lists
             inputs = copy.deepcopy(self.configSpec.inputs)
             relationship = isinstance(self.target, RelationshipInstance)
             if relationship:
-                target = self.target.target
+                target = self.target.target  # type: ignore  # checked with isinstance above
             else:
                 target = self.target
             HOST = (target.parent or target).attributes
@@ -372,24 +419,24 @@ class TaskView:
                 or {},
             )
             if relationship:
-                vars["SOURCE"] = self.target.source.attributes
+                vars["SOURCE"] = self.target.source.attributes  # type: ignore
                 vars["TARGET"] = target.attributes
             # expose inputs lazily to allow self-referencee
-            ctx = RefContext(self.target, vars, task=self)
+            ctx = RefContext(self.target, vars, task=self, strict=not self._rendering)
             if self.configSpec.artifact and self.configSpec.artifact.base_dir:
                 ctx.base_dir = self.configSpec.artifact.base_dir
             self._inputs = ResultsMap(inputs, ctx)
         return self._inputs
 
     @property
-    def vars(self):
+    def vars(self) -> dict:
         """
         A dictionary of the same variables that are available to expressions when evaluating inputs.
         """
         return self.inputs.context.vars
 
     @staticmethod
-    def _get_connection(source, target, seen):
+    def _get_connection(source: NodeInstance, target: Optional[NodeInstance], seen: dict) -> None:
         """
         Find the requirements on source that match the target
         If source is root, requirements will be the connections that are the default_for the target.
@@ -400,7 +447,7 @@ class TaskView:
             if id(rel) not in seen:
                 seen[id(rel)] = rel
 
-    def _get_connections(self):
+    def _get_connections(self) -> _ConnectionsMap:
         """
         Build a dictionary of connections by looking for instances that the task's implementation
         might to connect to (transitively following the target's hostedOn relationship)
@@ -408,7 +455,7 @@ class TaskView:
         Then add any default connections, prioritizing default connections to those instances.
         (Connections that explicity set a ``default_for`` key that matches those instances.)
         """
-        seen = {}
+        seen: Dict[int, Any] = {}
         for parent in self.target.ancestors:
             if parent is self.target.root:
                 break
@@ -424,7 +471,7 @@ class TaskView:
         )
         return connections
 
-    def _find_relationship_env_vars(self):
+    def _find_relationship_env_vars(self) -> dict:
         """
         Collect any environment variables set by the connections returned by ``_get_connections()``.
 
@@ -440,7 +487,7 @@ class TaskView:
 
         return env
 
-    def get_environment(self, addOnly):
+    def get_environment(self, addOnly: bool) -> dict:
         """Return a dictionary of environment variables applicable to this task.
 
         Args:
@@ -477,7 +524,7 @@ class TaskView:
         env = filter_env(rules, env, addOnly=addOnly)
 
         # add the variables required by TOSCA 1.3 spec
-        targets = []
+        targets: List[str] = []
         if isinstance(self.target, RelationshipInstance):
             targets = [
                 c.tosca_id
@@ -502,11 +549,11 @@ class TaskView:
             )
         return env
 
-    def get_settings(self):
+    def get_settings(self) -> dict:
         return dict(
-            verbose=self.verbose,
+            verbose=self.verbose,  # type: ignore
             name=self.configSpec.name,
-            dryrun=self.dry_run,
+            dryrun=self.dry_run,  # type: ignore
             workflow=self.configSpec.workflow,
             operation=self.configSpec.operation,
             timeout=self.configSpec.timeout,
@@ -515,7 +562,7 @@ class TaskView:
             cwd=self.cwd,
         )
 
-    def find_connection(self, target, relation="tosca.relationships.ConnectsTo"):
+    def find_connection(self, target: NodeInstance, relation: str="tosca.relationships.ConnectsTo") -> Any:
         connection = self.query(
             f"$OPERATION_HOST::.requirements::*[.type={relation}][.target=$target]",
             vars=dict(target=target),
@@ -528,7 +575,7 @@ class TaskView:
                 connection = endpoints[0]
         return connection
 
-    def sensitive(self, value):
+    def sensitive(self, value: object) -> Union[sensitive, object]:
         """Mark the given value as sensitive. Sensitive values will be encrypted or redacted when outputed.
 
         Returns:
@@ -537,10 +584,10 @@ class TaskView:
         """
         return wrap_sensitive_value(value)
 
-    def add_message(self, message):
+    def add_message(self, message: object) -> None:
         self.messages.append(message)
 
-    def find_instance(self, name):
+    def find_instance(self, name: str) -> NodeInstance:
         return self._manifest.get_root_resource().find_instance_or_external(name)
 
     # XXX
@@ -555,14 +602,14 @@ class TaskView:
 
     def done(
         self,
-        success=None,
-        modified=None,
-        status=None,
-        result=None,
-        outputs=None,
-        captureException=None,
-    ):
-        """`run()` should call this method and yield its return value before terminating.
+        success: Optional[bool]=None,
+        modified: Optional[bool]=None,
+        status: Optional[Status]=None,
+        result: Optional[Union[dict, str]]=None,
+        outputs: Optional[dict]=None,  # so the docstring says dict, but ConfiguratorResult
+        captureException: Optional[object]=None,
+    ) -> ConfiguratorResult:
+        """:py:meth:`unfurl.configurator.Configurator.run` should call this method and yield its return value before terminating.
 
         >>> yield task.done(True)
 
@@ -585,10 +632,20 @@ class TaskView:
             status = modified
             modified = True
 
-        kw = dict(result=result, outputs=outputs)
+        kw = dict(result=result, outputs=outputs)  # type: kwType
         if captureException is not None:
             logLevel = logging.DEBUG if success else logging.ERROR
-            kw["exception"] = UnfurlTaskError(self, captureException, logLevel)
+            kw["exception"] = UnfurlTaskError(self, captureException, logLevel)  # type: ignore
+
+        # Typechecking
+        class kwTypeBase(TypedDict):
+            result: Optional[dict]
+            outputs: Optional[dict]
+
+        class kwType(kwTypeBase, total=False):
+            exception: UnfurlTaskError
+
+        kw = cast(kwType, kw)
 
         return ConfiguratorResult(success, modified, status, **kw)
 
@@ -598,16 +655,16 @@ class TaskView:
     # other configurations maybe modify those changes, triggering a configuration change
     def query(
         self,
-        query,
-        dependency=False,
-        name=None,
-        required=False,
-        wantList=False,
-        resolveExternal=True,
-        strict=True,
-        vars=None,
-        throw=False,
-    ):
+        query: str,
+        dependency: bool=False,
+        name: Optional[str]=None,
+        required: bool=False,
+        wantList: bool=False,
+        resolveExternal: bool=True,
+        strict: bool=True,
+        vars: Optional[dict]=None,
+        throw: bool=False,
+    ) -> Optional[Ref]:
         # XXX pass resolveExternal to context?
         try:
             result = Ref(query, vars=vars).resolve(
@@ -629,14 +686,14 @@ class TaskView:
 
     def add_dependency(
         self,
-        expr,
-        expected=None,
-        schema=None,
-        name=None,
-        required=True,
-        wantList=False,
-        target=None,
-    ):
+        expr: Union[str, Mapping],
+        expected: Optional[Union[list, ResultsList, Result]]=None,
+        schema: Optional[Mapping]=None,
+        name: Optional[str]=None,
+        required: bool=True,
+        wantList: bool=False,
+        target: Optional[NodeInstance]=None,
+    ) -> "Dependency":
         getter = getattr(expr, "as_ref", None)
         if getter:
             # expr is a configuration or resource or ExternalValue
@@ -654,7 +711,7 @@ class TaskView:
         self._dependenciesChanged = True
         return dependency
 
-    def remove_dependency(self, name):
+    def remove_dependency(self, name: str) -> Optional["Dependency"]:
         for i, dep in enumerate(self.dependencies):
             if dep.name == name:
                 self.dependencies.pop(i)
@@ -668,9 +725,14 @@ class TaskView:
     #     return self._manifest.loadConfigSpec(name, configSpec)
 
     def create_sub_task(
-        self, operation=None, resource=None, inputs=None, persist=False, required=None
-    ):
-        """Create a subtask that will be executed if yielded by `run()`
+        self,
+        operation: Optional[str]=None,
+        resource: Optional[NodeInstance]=None,
+        inputs: Optional[dict]=None,
+        persist: bool=False,
+        required: Optional[bool]=None
+    ) -> Optional[TaskRequest]:
+        """Create a subtask that will be executed if yielded by :py:meth:`unfurl.configurator.Configurator.run`
 
         Args:
           operation (str): The operation call (like ``interface.operation``)
@@ -689,7 +751,7 @@ class TaskView:
             operation = f"{self.configSpec.interface}.{self.configSpec.operation}"
         if isinstance(operation, six.string_types):
             taskRequest = create_task_request(
-                self.job.jobOptions,
+                self.job.jobOptions,  # type: ignore
                 operation,
                 resource,
                 "subtask: " + self.configSpec.name,
@@ -715,7 +777,7 @@ class TaskView:
         # otherwise operation should be a ConfigurationSpec
         return TaskRequest(operation, resource, "subtask", persist, required)
 
-    def _update_instance(self, existingResource, resourceSpec):
+    def _update_instance(self, existingResource: NodeInstance, resourceSpec: Mapping) -> bool:
         from .manifest import Manifest
 
         updated = False
@@ -726,25 +788,27 @@ class TaskView:
             # XXX track all status attributes (esp. state and created) and remove this hack
             operational = Manifest.load_status(resourceSpec)
             if operational.local_status is not None:
-                existingResource.local_status = operational.local_status
+                existingResource.local_status = operational.local_status  # type: ignore
             if operational.state is not None:
-                existingResource.state = operational.state
+                existingResource.state = operational.state  # type: ignore
             updated = True
 
         attributes = resourceSpec.get("attributes")
         if attributes:
-            for key, value in map_value(attributes, existingResource).items():
-                existingResource.attributes[key] = value
+            for key, value in attributes.items():
+                existingResource.attributes[key] = map_value(value, existingResource)
                 self.logger.debug(
-                    "setting attribute %s with %s on %s",
+                    "setting attribute %s on %s with %s",
                     key,
-                    value,
                     existingResource.name,
+                    truncate_str(value)
                 )
             updated = True
         return updated
 
-    def _parse_instances_tpl(self, instances):
+    def _parse_instances_tpl(self,
+        instances: Union[str, Mapping, MutableSequence]
+    ) -> Union[Tuple[None, UnfurlTaskError], Tuple[MutableSequence, None]]:
         if isinstance(instances, six.string_types):
             try:
                 instances = yaml.load(instances)
@@ -764,7 +828,7 @@ class TaskView:
 
     # # XXX how can we explicitly associate relations with target resources etc.?
     # # through capability attributes and dependencies/relationship attributes
-    def update_instances(self, instances):
+    def update_instances(self, instances: Union[str, List]) -> Union[Tuple[JobRequest, List], Tuple[None, List]]:
         """Notify Unfurl of new or changes to instances made while the configurator was running.
 
         Operational status indicates if the instance currently exists or not.
@@ -791,7 +855,7 @@ class TaskView:
           :class:`JobRequest`: To run the job based on the supplied spec
               immediately, yield the returned JobRequest.
         """
-        instances, err = self._parse_instances_tpl(instances)
+        instances, err = self._parse_instances_tpl(instances)  # type: ignore
         if err:
             return None, [err]
 
@@ -841,12 +905,12 @@ class TaskView:
             self.logger.info("add resources %s", newResources)
 
             jobRequest = JobRequest(newResources, errors)
-            if self.job:
-                self.job.jobRequestQueue.append(jobRequest)
+            if self.job:  # type: ignore
+                self.job.jobRequestQueue.append(jobRequest)  # type: ignore
             return jobRequest, errors
         return None, errors
 
-    def set_work_folder(self, location="operation", preserve=None) -> WorkFolder:
+    def set_work_folder(self, location: str="operation", preserve: Optional[bool]=None) -> WorkFolder:
         if location in self._workFolders:
             return self._workFolders[location]
         if preserve is None:
@@ -859,7 +923,7 @@ class TaskView:
         #     self, location, preserve
         # )
 
-    def get_work_folder(self, location=None):
+    def get_work_folder(self, location: Optional[str]=None) -> WorkFolder:
         # return self.job.getFolder(self, location)
         if location is None:
             # XXX error if there is more than one?
@@ -867,19 +931,19 @@ class TaskView:
         else:
             return self._workFolders[location]
 
-    def discard_work_folders(self):
+    def discard_work_folders(self) -> None:
         while self._workFolders:
             _, wf = self._workFolders.popitem()
             wf.discard()
 
-    def fail_work_folders(self):
+    def fail_work_folders(self) -> None:
         while self._workFolders:
             _, wf = self._workFolders.popitem()
             wf.failed()
 
-    def apply_work_folders(self, *names):
+    def apply_work_folders(self, *names: str) -> None:
         if not names:  # no args were passed, apply them all
-            names = self._workFolders.keys()
+            names = self._workFolders.keys()  # type: ignore
         for name in names:
             wf = self._workFolders.get(name)
             if wf:
@@ -897,14 +961,14 @@ class Dependency(Operational):
 
     def __init__(
         self,
-        expr,
-        expected=None,
-        schema=None,
-        name=None,
-        required=False,
-        wantList=False,
-        target=None,
-    ):
+        expr: Union[str, dict],
+        expected: Optional[Union[ResultsList, list, Result, None]]=None,
+        schema: Optional[Mapping]=None,
+        name: Optional[str]=None,
+        required: bool=False,
+        wantList: bool=False,
+        target: Optional[NodeInstance]=None,
+    ) -> None:
         """
         if schema is not None, validate the result using schema
         if expected is not None, test that result equals expected
@@ -921,17 +985,19 @@ class Dependency(Operational):
         self.target = target
 
     @property
-    def local_status(self):
-        if self.target and self.target is not self.target.root:
-            return self.target.status
+    def local_status(self) -> Status:
+        if self.target and self.target is not self.target.root and not self.target.is_computed():
+            # (only care about local status of instances with live attribute, not their full operational status)
+            # (reduces circular dependencies)
+            return self.target.local_status  # type: ignore  # mypy has trouble with @property calls
         else:  # the root has inputs which don't have operational status
             return Status.ok
 
     @property
-    def priority(self):
+    def priority(self) -> Priority:
         return Priority.required if self._required else Priority.optional
 
-    def refresh(self, config):
+    def refresh(self, config: "ConfigTask") -> None:
         if self.expected is not None:
             changeId = config.changeId
             context = RefContext(
@@ -942,7 +1008,7 @@ class Dependency(Operational):
             self.expected = result
 
     @staticmethod
-    def has_value_changed(value, changeset):
+    def has_value_changed(value: Union[Results, Mapping, MutableSequence, tuple, ChangeAware, Any], changeset: ChangeRecord) -> bool:
         if isinstance(value, Results):
             return Dependency.has_value_changed(value._attributes, changeset)
         elif isinstance(value, Mapping):
@@ -955,8 +1021,9 @@ class Dependency(Operational):
             return value.has_changed(changeset)
         else:
             return False
+        return False  # assuming this is what is meant by the function
 
-    def has_changed(self, config):
+    def has_changed(self, config: "ConfigTask") -> bool:
         changeId = config.changeId
         context = RefContext(config.target, dict(val=self.expected, changeId=changeId))
         result = Ref(self.expr).resolve_one(context)  # resolve(context, self.wantList)
