@@ -89,6 +89,8 @@ class TemplateConfigurator(Configurator):
         return runResult
 
     def done(self, task, **kw):
+        # this is called by derived classes like ShellConfigurator to allow the user
+        # to override the default logic for updating the state and status of a task.
         done = task.inputs.get_copy("done", {})
         if done:
             task.logger.trace("evaluated done template with %s", done)
@@ -103,7 +105,7 @@ class TemplateConfigurator(Configurator):
                 done["result"] = {"run": runResult}
             else:
                 done["result"] = runResult
-        if self.process_result_template(task, done.get("result")):
+        if self.process_result_template(task, done.get("result") or {}):
             done["success"] = False  # returned errors
         yield task.done(**done)
 
