@@ -2,14 +2,17 @@
 # SPDX-License-Identifier: MIT
 from __future__ import absolute_import
 import codecs
+import re
 from ..configurator import Configurator, Status
 from ..runtime import RelationshipInstance
 from .ansible import AnsibleConfigurator
 import json
-from ansible.module_utils.k8s.common import K8sAnsibleMixin
+from ansible_collections.kubernetes.core.plugins.module_utils.common import (
+    K8sAnsibleMixin,
+)
 
 
-def _getConnectionConfig(instance):
+def _get_connection_config(instance, cluster):
     # see https://docs.ansible.com/ansible/latest/modules/k8s_module.html#k8s-module
     #  for connection settings
     if not instance:
@@ -17,15 +20,37 @@ def _getConnectionConfig(instance):
     connect = {}
     if isinstance(instance, RelationshipInstance):
         connect = instance.attributes
-        endpoint = instance.parent.attributes
+        # note: endpoint will be root if instance is a default connection
+        if instance.parent is not instance.root:
+            endpoint = instance.parent.attributes
+        else:
+            endpoint = {}
     else:
         endpoint = instance.attributes
 
     connection = {}
-    if "host" in endpoint:  # endpoint capability only
-        connection["host"] = endpoint["host"]
+    # cluster only:
+    if cluster.attributes.get("api_server"):
+        connection["host"] = "https://" + cluster.attributes["api_server"]
 
-    map1 = {"KUBECONFIG": "kubeconfig", "context": "context", "secure": "verify_ssl"}
+    # endpoint capability only
+    protocol = endpoint.get("protocol")
+    ip_address = endpoint.get("ip_address")
+    port = endpoint.get("port")
+    if ip_address and port:
+        connection["host"] = f"{protocol}://{ip_address}:{port}"
+
+    # connection only:
+    api_server = connect.get("api_server")
+    if api_server:
+        connection["host"] = f"https://{api_server}"
+
+    map1 = {
+        "KUBECONFIG": "kubeconfig",
+        "context": "context",
+        "secure": "verify_ssl",
+        "namespace": "namespace",
+    }
     # relationship overrides capability
     for attributes in [endpoint, connect]:
         for key, value in map1.items():
@@ -45,29 +70,32 @@ def _getConnectionConfig(instance):
     return connection
 
 
-def _getConnection(task, cluster):
-    instance = task.findConnection(
+def _get_connection(task, cluster):
+    instance = task.find_connection(
         cluster, relation="unfurl.relationships.ConnectsTo.K8sCluster"
     )
-    return _getConnectionConfig(instance)
+    return _get_connection_config(instance, cluster)
 
 
 class ClusterConfigurator(Configurator):
     @staticmethod
-    def _getHost(connectionConfig):
+    def _get_host(connectionConfig):
         client = K8sAnsibleMixin().get_api_client(**connectionConfig)
-        return client.configuration.host
+        url = client.configuration.host
+        if url:
+            return re.sub("^https?://", "", url)
+        return url
 
-    def canRun(self, task):
+    def can_run(self, task):
         if task.configSpec.operation not in ["check", "discover"]:
             return "Configurator can't perform this operation (only supports check and discover)"
         return True
 
     def run(self, task):
         cluster = task.target
-        connectionConfig = _getConnection(task, cluster)
+        connectionConfig = _get_connection(task, cluster)
         try:
-            cluster.attributes["apiServer"] = self._getHost(connectionConfig)
+            cluster.attributes["api_server"] = self._get_host(connectionConfig)
         except:
             yield task.done(
                 False,
@@ -79,28 +107,28 @@ class ClusterConfigurator(Configurator):
 
 
 class ResourceConfigurator(AnsibleConfigurator):
-    def getGenerator(self, task):
-        if task.dryRun:
-            return self.dryRun(task)
+    def get_generator(self, task):
+        if task.dry_run:
+            return self.dry_run(task)
         else:
             return self.run(task)
 
-    def dryRun(self, task):
+    def dry_run(self, task):
         # XXX don't use print()
         print("generating playbook")
         # print(self.findPlaybook(task))
         # print(self.findPlaybook(task))
-        print(json.dumps(self.findPlaybook(task), indent=4))
+        print(json.dumps(self.find_playbook(task), indent=4))
         yield task.done(True)
 
-    def _getConnection(self, task):
+    def _get_connection(self, task):
         # get the cluster that the target resource is hosted on
         cluster = task.query("[.type=unfurl.nodes.K8sCluster]")
         if not cluster:
             return {}
-        return _getConnection(task, cluster)
+        return _get_connection(task, cluster)
 
-    def makeSecret(self, data):
+    def make_secret(self, data):
         # base64 adds trailing \n so strip it out
         return dict(
             type="Opaque",
@@ -112,27 +140,27 @@ class ResourceConfigurator(AnsibleConfigurator):
             },
         )
 
-    def getDefinition(self, task):
-        if task.target.template.isCompatibleType("unfurl.nodes.K8sNamespace"):
+    def get_definition(self, task):
+        if task.target.template.is_compatible_type("unfurl.nodes.K8sNamespace"):
             return dict(apiVersion="v1", kind="Namespace")
 
         if "definition" in task.target.attributes:
-            definition = task.target.attributes.getCopy("definition")
+            definition = task.target.attributes.get_copy("definition") or {}
         else:
-            definition = task.target.attributes.getCopy("apiResource", {})
+            definition = task.target.attributes.get_copy("apiResource") or {}
 
-        if not definition and task.target.template.isCompatibleType(
+        if not definition and task.target.template.is_compatible_type(
             "unfurl.nodes.K8sSecretResource"
         ):
-            return self.makeSecret(task.target.attributes.get("data", {}))
+            return self.make_secret(task.target.attributes.get("data", {}))
         else:
             # XXX if definition is string: parse
             # get copy so subsequent modifications dont affect the definition
             return definition
 
-    def updateMetadata(self, definition, task):
+    def update_metadata(self, definition, task):
         namespace = None
-        if task.target.parent.template.isCompatibleType("unfurl.nodes.K8sNamespace"):
+        if task.target.parent.template.is_compatible_type("unfurl.nodes.K8sNamespace"):
             namespace = task.target.parent.attributes["name"]
         md = definition.setdefault("metadata", {})
         if namespace and "namespace" not in md:
@@ -146,12 +174,12 @@ class ResourceConfigurator(AnsibleConfigurator):
         else:
             md["name"] = name
 
-    def findPlaybook(self, task):
-        definition = self.getDefinition(task)
-        self.updateMetadata(definition, task)
+    def find_playbook(self, task):
+        definition = self.get_definition(task)
+        self.update_metadata(definition, task)
         delete = task.configSpec.operation in ["Standard.delete", "delete"]
         state = "absent" if delete else "present"
-        connectionConfig = self._getConnection(task)
+        connectionConfig = self._get_connection(task)
         moduleSpec = dict(state=state, **connectionConfig)
         if task.configSpec.operation in ["check", "discover"]:
             moduleSpec["kind"] = definition.get("kind", "")
@@ -160,9 +188,9 @@ class ResourceConfigurator(AnsibleConfigurator):
                 moduleSpec["namespace"] = definition["metadata"]["namespace"]
         else:
             moduleSpec["resource_definition"] = definition
-        return [dict(k8s=moduleSpec)]
+        return [{"community.kubernetes.k8s": moduleSpec}]
 
-    def processResult(self, task, result):
+    def process_result(self, task, result):
         # overrides super.processResult
         resource = result.result.get("result")
         task.target.attributes["apiResource"] = resource
@@ -184,6 +212,6 @@ class ResourceConfigurator(AnsibleConfigurator):
                 result.status = states.get(status, Status.unknown)
         return result
 
-    def getResultKeys(self, task, results):
+    def get_result_keys(self, task, results):
         # save first time even if it hasn't changed
         return ["result"]  # also "method", "diff", invocation
