@@ -15,10 +15,11 @@ import re
 def _get_env(env, verbose, dataDir):
     env["TF_IN_AUTOMATION"] = "1"
     env["TF_INPUT"] = "0"
-    # terraform currently only supports TF_LOG=TRACE
+    # see https://www.terraform.io/plugin/log/managing
     # env["TF_LOG"] = "ERROR WARN INFO DEBUG TRACE".split()[verbose + 1]
     if verbose > 0:
-        env["TF_LOG"] = "DEBUG"
+        # providers can be very verbose, don't set them to debug
+        env["TF_LOG_CORE"] = "DEBUG"
 
     # note: modules with relative paths get confused .terraform isn't child of the config dir
     # contains modules/modules.json and plugins/plugins.json:
@@ -94,11 +95,16 @@ module "main" {
 
 def generate_main(relpath, tfvars, outputs):
     # if tfvars are hcl:
-    # XXX don't mark every output as sensitive
+    # XXX check if outputs are sensitive
     if isinstance(tfvars, six.string_types):
         output = ""
         for name in outputs:
-            output += f'output "{name}" {{\n value = module.main.{name}\n sensitive=true\n }}\n'
+            sensitive = False
+            if sensitive:
+                sensitive_str = "sensitive=true\n"
+            else:
+                sensitive_str = ""
+            output += f'output "{name}" {{\n value = module.main.{name}\n {sensitive_str} }}\n'
         return "main.tmp.tf", _main_tf_template % (relpath, tfvars, output)
     else:
         # place tfvars in the module block:
@@ -108,7 +114,8 @@ def generate_main(relpath, tfvars, outputs):
         if outputs:
             output = root["output"] = {}
             for name in outputs:
-                output[name] = dict(value=f"${{module.main.{name}}}", sensitive=True)
+                sensitive = False
+                output[name] = dict(value=f"${{module.main.{name}}}", sensitive=sensitive)
         return "main.tmp.tf.json", root
 
 
@@ -151,9 +158,7 @@ class TerraformConfigurator(ShellConfigurator):
         ]
         # folder is "tasks" WorkFolder
         cwd = folder.cwd
-        lock_file = get_path(
-            task.inputs.context, ".terraform.lock.hcl", Folders.artifacts
-        )
+        lock_file = task.set_work_folder(Folders.artifacts).permanent_path(".terraform.lock.hcl", False)
         if os.path.exists(lock_file):
             folder.copy_from(lock_file)
 
@@ -199,14 +204,15 @@ class TerraformConfigurator(ShellConfigurator):
         # generated tf.json get written to as main.unfurl.tmp.tf.json
         write_vars = True
         contents = None
-        main = task.inputs.get_copy("main")
-        if not main:
+        if "main" not in task.inputs:
             main = get_path(task.inputs.context, "", "spec.home")
             if not os.path.exists(main):
                 raise UnfurlTaskError(
                     task,
                     f'Input parameter "main" not specifed and default terraform module directory does not exist at "{main}"',
                 )
+        else:
+            main = task.inputs.get_copy("main")
         if isinstance(main, six.string_types):
             if "\n" in main:
                 # assume its HCL and not a path
@@ -268,7 +274,7 @@ class TerraformConfigurator(ShellConfigurator):
         return None
 
     def _get_workfolder_name(self, task):
-        return task.inputs.get("stateLocation") or Folders.secrets
+        return task.inputs.get("stateLocation") or Folders.artifacts # XXX global option for secrets
 
     def _prepare_state(self, task, cwd):
         # the terraform state file is associate with the current instance
@@ -277,12 +283,19 @@ class TerraformConfigurator(ShellConfigurator):
         folderName = self._get_workfolder_name(task)
         if folderName == "remote":  # don't use local state file
             return ""
-        yamlPath = get_path(task.inputs.context, "terraform.tfstate.yaml", folderName)
+        yamlPath = task.set_work_folder(folderName).permanent_path("terraform.tfstate.yaml", False)
         if os.path.exists(yamlPath):
+            task.logger.debug(
+                "Found existing terraform.tfstate file at %s", yamlPath
+            )
             # if exists in home, load and write out state file as json
             with open(yamlPath, "r") as f:
                 state = task._manifest.yaml.load(f.read())
             cwd.write_file(state, "terraform.tfstate")
+        else:
+            task.logger.debug(
+                "Couldn't find terraform.tfstate file at %s", yamlPath
+            )
         return "terraform.tfstate"
 
     def _get_plan_path(self, task, cwd):
@@ -435,10 +448,14 @@ class TerraformConfigurator(ShellConfigurator):
             or "Creating..." in result.stdout
             or "Destroying..." in result.stdout
         )
-        yield task.done(
-            success, modified, status, result=result.__dict__, outputs=outputs
-        )
-
+        yield self.done(
+                task,
+                success=success,
+                modified=modified,
+                status=status,
+                result=result.__dict__,
+                outputs=outputs,
+            )
 
 # XXX implement discover:
 # terraform import -allow-missing-config {type.name}
