@@ -15,21 +15,35 @@ eval_ref() given expression (string or dictionary) return list of Result
 Expr.resolve() given expression string, return list of Result
 Results._map_value same as map_value but with lazily evaluation
 """
+from typing import Any, Dict, List, Optional, Union, cast, TYPE_CHECKING
 import six
 import re
 import operator
+import sys
 import collections
 from collections.abc import Mapping, MutableSequence
 from ruamel.yaml.comments import CommentedMap
+
+from unfurl.logs import UnfurlLogger
+
 from .util import validate_schema, UnfurlError, assert_form
-from .result import ResultsList, Result, Results, ExternalValue, ResourceRef
+from .result import ResultsList, Result, Results, ExternalValue, ResourceRef, ResultsMap
 
 import logging
 
-logger = logging.getLogger("unfurl.eval")
+if TYPE_CHECKING:
+    from .configurator import TaskView
+    from .runtime import NodeInstance
 
 
-def map_value(value, resourceOrCxt, applyTemplates=True):
+logger = cast(UnfurlLogger, logging.getLogger("unfurl.eval"))
+
+
+def map_value(
+    value: Any,
+    resourceOrCxt: Union["RefContext", "NodeInstance"],
+    applyTemplates: bool=True
+    ) -> Optional[Union[ResultsList, Result, List[Result]]]:
     if not isinstance(resourceOrCxt, RefContext):
         resourceOrCxt = RefContext(resourceOrCxt)
     return _map_value(value, resourceOrCxt, False, applyTemplates)
@@ -74,7 +88,11 @@ class _Tracker:
         self.count -= 1
         return self.count
 
-    def addReference(self, ref, result):
+    def add_ref_reference(self, ref: "Ref", result: ResultsList):
+        if self.count > 0:
+            self.referenced.append([ref, result])
+
+    def add_result_reference(self, ref: Union[str, None], result: Result):
         if self.count > 0:
             self.referenced.append([ref, result])
 
@@ -104,14 +122,14 @@ class RefContext:
 
     def __init__(
         self,
-        currentResource,
-        vars=None,
-        wantList=False,
-        resolveExternal=False,
-        trace=None,
-        strict=_defaultStrictness,
-        task=None,
-    ):
+        currentResource: "NodeInstance",
+        vars: Optional[dict]=None,
+        wantList: bool=False,
+        resolveExternal: bool=False,
+        trace: Optional[int]=None,
+        strict: bool=_defaultStrictness,
+        task: "TaskView"=None,
+    ) -> None:
         self.vars = vars or {}
         # the original context:
         self.currentResource = currentResource
@@ -128,7 +146,14 @@ class RefContext:
         self.referenced = _Tracker()
         self.task = task
 
-    def copy(self, resource=None, vars=None, wantList=None, trace=0, strict=None):
+    def copy(
+        self,
+        resource: Optional["NodeInstance"]=None,
+        vars: dict=None,
+        wantList: bool=None,
+        trace:  int=0,
+        strict: bool=None
+        ) -> "RefContext":
         copy = RefContext(
             resource or self.currentResource,
             self.vars,
@@ -151,22 +176,26 @@ class RefContext:
         copy.task = self.task
         return copy
 
-    def trace(self, *msg):
+    def trace(self, *msg: Any) -> None:
         if self._trace:
-            logger.trace(f"{' '.join(str(a) for a in msg)} (ctx: {self._lastResource})")
+            log = logger.info if self._trace == 2 else logger.trace
+            log(f"{' '.join(str(a) for a in msg)} (ctx: {self._lastResource})")
 
-    def add_external_reference(self, external):
+    def add_external_reference(self, external: ExternalValue) -> Result:
         result = Result(external)
-        self.referenced.addReference(None, result)
+        self.referenced.add_result_reference(None, result)
         return result
 
-    def add_reference(self, ref, result):
-        self.referenced.addReference(ref, result)
+    def add_ref_reference(self, ref: "Ref", result: ResultsList) -> None:
+        self.referenced.add_ref_reference(ref, result)
 
-    def resolve_var(self, key):
+    def add_result_reference(self, ref: str, result: Result) -> None:
+        self.referenced.add_result_reference(ref, result)
+
+    def resolve_var(self, key: str) -> Optional[Union[Results, ResultsList, List[Result], ResultsMap]]:
         return self._resolve_var(key[1:]).resolved
 
-    def _resolve_var(self, key):
+    def _resolve_var(self, key: str) -> Result:
         value = self.vars[key]
         if isinstance(value, Result):
             return value
@@ -178,16 +207,21 @@ class RefContext:
             self.vars[key] = val
             return val
 
-    def resolve_reference(self, key):
+    def resolve_reference(self, key: str) -> Union[Results, ResultsList, List[Result], ResultsMap]:
         val = self._resolve_var(key)
-        self.add_reference(key, val)
+        self.add_result_reference(key, val)
         assert not isinstance(val.resolved, Result)
         return val.resolved
 
-    def query(self, expr, vars=None, wantList=False):
+    def query(
+        self,
+        expr: Mapping,
+        vars: Optional[dict]=None,
+        wantList: Union[bool, str]=False
+    ) -> Optional[Union[ResultsList, Result, List[Result]]]:
         return Ref(expr, vars).resolve(self, wantList)
 
-    def __getstate__(self):
+    def __getstate__(self) -> Dict[str, Any]:
         # Remove the unpicklable entries.
         state = self.__dict__.copy()
         state["templar"] = None
@@ -195,7 +229,7 @@ class RefContext:
         del state["referenced"]
         return state
 
-    def __setstate__(self, d):
+    def __setstate__(self, d: dict) -> None:
         self.__dict__ = d
         self.referenced = _Tracker()
 
@@ -250,24 +284,30 @@ class Expr:
 class Ref:
     """A Ref objects describes a path to metadata associated with a resource."""
 
-    def __init__(self, exp, vars=None):
+    def __init__(self, exp: Union[str, Mapping], vars: dict=None, trace: Optional[int]=None) -> None:
         self.vars = {"true": True, "false": False, "null": None}
 
         self.foreach = None
-        self.trace = 0
+        trace = RefContext.DefaultTraceLevel if trace is None else trace
+        self.trace = trace
         if isinstance(exp, Mapping):
             keys = list(exp)
             if keys and keys[0] not in _FuncsTop:
                 self.vars.update(exp.get("vars", {}))
                 self.foreach = exp.get("foreach", exp.get("select"))
-                self.trace = exp.get("trace", 0)
+                self.trace = exp.get("trace", trace)
                 exp = exp.get("eval", exp.get("ref", exp))
 
         if vars:
             self.vars.update(vars)
         self.source = exp
 
-    def resolve(self, ctx, wantList=True, strict=_defaultStrictness):
+    def resolve(
+        self,
+        ctx: RefContext,
+        wantList: Union[bool, str]=True,
+        strict: bool=_defaultStrictness
+    ) -> Optional[Union[ResultsList, Result, List[Result]]]:
         """
         If wantList=True (default) returns a ResultList of matches
         Note that values in the list can be a list or None
@@ -280,7 +320,9 @@ class Ref:
         base_dir = getattr(self.source, "base_dir", None)
         if base_dir:
             ctx.base_dir = base_dir
-
+        # $start is set in eval_ref but the user use that to override currentResource
+        if 'start' in self.vars:
+            ctx.currentResource = ctx.resolve_var('$start')
         ctx.trace(
             f"Ref.resolve(wantList={wantList}) start strict {ctx.strict}",
             self.source,
@@ -291,7 +333,7 @@ class Ref:
             results = for_each(self.foreach, results, ctx)
         assert not isinstance(results, ResultsList), results
         results = ResultsList(results, ctx)
-        ctx.add_reference(self, results)
+        ctx.add_ref_reference(self, results)
         ctx.trace(f"Ref.resolve(wantList={wantList}) results", self.source, results)
         if wantList and not wantList == "result":
             return results
@@ -309,7 +351,7 @@ class Ref:
                 else:
                     return list(results)
 
-    def resolve_one(self, ctx, strict=_defaultStrictness):
+    def resolve_one(self, ctx: RefContext, strict: bool=_defaultStrictness) -> Optional[Union[ResultsList, Result, List[Result]]]:
         """
         If no match return None
         If more than one match return a list of matches
@@ -322,7 +364,7 @@ class Ref:
         return self.resolve(ctx, False, strict)
 
     @staticmethod
-    def is_ref(value):
+    def is_ref(value: Union[Mapping, "Ref"]) -> bool:
         if isinstance(value, Mapping):
             if not value:
                 return False
@@ -493,8 +535,11 @@ def set_eval_func(name, val, topLevel=False):
     if topLevel:
         _FuncsTop.append(name)
 
-
-def eval_ref(val, ctx, top=False):
+if sys.version_info >= (3, 9):
+    MutableSequence_Result = MutableSequence[Result]
+else:
+    MutableSequence_Result = MutableSequence
+def eval_ref(val: Union[Mapping, str], ctx: RefContext, top: bool=False) -> MutableSequence_Result:
     "val is assumed to be an expression, evaluate and return a list of Result"
     from .support import is_template, apply_template
 
@@ -509,10 +554,10 @@ def eval_ref(val, ctx, top=False):
             func = _Funcs.get(key)
             if func:
                 args = val[key]
-                ctx.kw = val
-                ctx.currentFunc = key
+                ctx.kw = val  # type: ignore
+                ctx.currentFunc = key  # type: ignore
                 if "var" in val:
-                    unexpected = "var"
+                    unexpected: Union[bool, str] = "var"
                 elif key != "foreach" and "foreach" in val:
                     unexpected = "foreach"
                 else:
