@@ -6,10 +6,9 @@ This module implements creating and cloning project and ensembles as well Unfurl
 import datetime
 import os
 import os.path
-import random
 import shutil
-import string
 import sys
+from typing import Any
 import uuid
 import logging
 from jinja2.loaders import FileSystemLoader
@@ -18,7 +17,7 @@ from pathlib import Path
 from . import DefaultNames, __version__, get_home_config_path, is_version_unreleased
 from .localenv import LocalEnv, Project, LocalConfig
 from .repo import GitRepo, Repo, is_url_or_git_path, split_git_url, commit_secrets
-from .util import UnfurlError
+from .util import UnfurlError, get_random_password
 from .yamlloader import make_yaml, make_vault_lib
 
 _templatePath = os.path.join(os.path.abspath(os.path.dirname(__file__)), "templates")
@@ -31,17 +30,6 @@ def rename_for_backup(dir):
     return new
 
 
-def get_random_password(count=12, prefix="uv", extra=None):
-    srandom = random.SystemRandom()
-    start = string.ascii_letters + string.digits
-    if extra is None:
-        extra = "%&()*+,-./:<>?=@^_`~"
-    source = string.ascii_letters + string.digits + extra
-    return prefix + "".join(
-        srandom.choice(source if i else start) for i in range(count)
-    )
-
-
 def _write_file(folder, filename, content):
     if not os.path.isdir(folder):
         os.makedirs(os.path.normpath(folder))
@@ -522,7 +510,7 @@ def _looks_like(path, name):
     return None
 
 
-def _get_ensemble_paths(sourcePath, sourceProject):
+def _get_ensemble_paths(sourcePath, sourceProject, want_init):
     """
     Returns either a pointer to the ensemble to clone
     or a dict of variables to pass to an ensemble template to create a new one
@@ -543,7 +531,7 @@ def _get_ensemble_paths(sourcePath, sourceProject):
 
     # we only support cloning TOSCA service templates if their names end in "service-template.yaml"
     isServiceTemplate = sourcePath.endswith(DefaultNames.ServiceTemplate)
-    if not isServiceTemplate:
+    if not want_init and not isServiceTemplate:
         try:
             localEnv = LocalEnv(relPath, project=sourceProject)
             sourceDir = sourceProject.get_relative_path(
@@ -645,9 +633,10 @@ class EnsembleBuilder:
         # source is a path into the project relative to the current directory
         source_path = os.path.join(self.source_project.projectRoot, self.source_path)
         self.templateVars = _get_ensemble_paths(
-            source_path,
-            self.source_project,
+            source_path, self.source_project, self.options.get("want_init")
         )
+        if self.options.get("use_deployment_blueprint"):
+            self.templateVars['deployment_blueprint'] = self.options['use_deployment_blueprint']
         (self.environment, self.shared_repo) = _get_context_and_shared_repo(
             self.dest_project, self.options
         )
@@ -740,11 +729,11 @@ class EnsembleBuilder:
         )
 
         templateVars = self.templateVars
-        if "localEnv" not in templateVars or self.options.get("want_init"):
+        if "localEnv" not in templateVars:
             # we found a template file to clone
             templateVars["inputs"] = self._get_inputs_template()
             localEnv, manifest = self._create_ensemble_from_template(
-                destProject, destDir, manifestName
+                destProject, destDir, manifestName,
             )
         else:
             # look for an ensemble at the given path or use the source project's default
@@ -776,10 +765,16 @@ class EnsembleBuilder:
             ".."
         ), f"{self.source_path} should be inside the project"
         if currentProject:
-            newrepo = currentProject.find_or_create_working_dir(
-                sourceProject.project_repoview.repo.url,
-            )
-            search = os.path.join(newrepo.working_dir, self.source_path)
+            repoURL = sourceProject.project_repoview.repo.url
+            if currentProject.find_git_repo(repoURL):
+                # if the repo has already been cloned into this project, just use that one
+                newrepo = currentProject.find_or_create_working_dir(repoURL)
+                search = os.path.join(newrepo.working_dir, self.source_path)
+            else:
+                # just register the source project's repo as a localRepository
+                currentProject.localConfig.register_project(sourceProject, save_project=False)
+                self.source_project = sourceProject
+                return sourceProject
         else:
             # clone the source project's git repo
             newrepo = sourceProject.project_repoview.repo.clone(dest_dir)
@@ -809,7 +804,7 @@ class EnsembleBuilder:
             Repo.create_working_dir(repoURL, destDir, revision)
 
         targetDir = os.path.join(destDir, filePath)
-        sourceProjectRoot = Project.find_path(targetDir)
+        sourceProjectRoot = Project.find_path(targetDir, destDir)
         self.logger.debug(f'cloned {self.input_source}" to "{destDir}"')
         if not sourceProjectRoot:
             raise UnfurlError(
@@ -827,7 +822,7 @@ class EnsembleBuilder:
         assert self.dest_project is None
         new_project = self.source_project is not existingSourceProject
         if existingDestProject:
-            #     set that as the dest_project
+            #  set that as the dest_project
             self.dest_project = existingDestProject
         else:
             # otherwise set source_project as the dest_project
@@ -908,7 +903,7 @@ class EnsembleBuilder:
         return f'Created new ensemble at "{os.path.abspath(destDir)}"'
 
 
-def clone(source, dest, ensemble_name=DefaultNames.EnsembleDirectory, **options):
+def clone(source: str, dest: str, ensemble_name: str=DefaultNames.EnsembleDirectory, **options: Any) -> str:
     """
     Clone the ``source`` ensemble to ``dest``. If ``dest`` isn't in a project, create one.
     ``source`` can point to an ensemble_template, a service_template, an existing ensemble
