@@ -17,7 +17,7 @@ from .runtime import (
 )
 from .util import UnfurlError, to_enum, sensitive_str, get_base_dir
 from .repo import RevisionManager, split_git_url, RepoView
-from .merge import patch_dict
+from .merge import merge_dicts
 from .yamlloader import YamlConfig, yaml, ImportResolver
 from .job import ConfigChange
 import toscaparser.imports
@@ -58,7 +58,9 @@ class Manifest(AttributeManager):
         repositories = {
             name: repo.repository.tpl for name, repo in self.repositories.items()
         }
-        self.tosca = self._load_spec(spec, self.path, repositories, more_spec, skip_validation)
+        self.tosca = self._load_spec(
+            spec, self.path, repositories, more_spec, skip_validation
+        )
         self.specDigest = self.get_spec_digest(spec)
 
     def _find_repo(self):
@@ -87,8 +89,10 @@ class Manifest(AttributeManager):
         toscaDef["tosca_definitions_version"] = TOSCA_VERSION
 
         if more_spec:
-            patch_dict(toscaDef, more_spec, True)
-
+            # don't merge individual templates
+            toscaDef = merge_dicts(
+                toscaDef, more_spec, replaceKeys=['node_templates', 'relationship_templates']
+            )
         if not isinstance(toscaDef, CommentedMap):
             toscaDef = CommentedMap(toscaDef.items())
         if getattr(toscaDef, "base_dir", None) and (
@@ -96,7 +100,9 @@ class Manifest(AttributeManager):
         ):
             # note: we only recorded the baseDir not the name of the included file
             path = toscaDef.base_dir
-        return ToscaSpec(toscaDef, spec, path, self.get_import_resolver(expand=True), skip_validation)
+        return ToscaSpec(
+            toscaDef, spec, path, self.get_import_resolver(expand=True), skip_validation
+        )
 
     def get_spec_digest(self, spec):
         m = hashlib.sha1()  # use same digest function as git
@@ -321,7 +327,7 @@ class Manifest(AttributeManager):
                 changerecord = self._get_last_change(operational)
                 template = self.load_template(templateName, changerecord)
         if template is None:
-            raise UnfurlError(f"missing resource template {templateName}")
+            raise UnfurlError(f"missing template definition for '{templateName}' while instantiating instance '{name}'")
         # logger.debug("creating instance for template %s: %s", templateName, template)
 
         # omit keys that match <<REDACTED>> so can we use the computed property
@@ -335,6 +341,9 @@ class Manifest(AttributeManager):
             instance.created = status["created"]
         if imported:
             self.imports.set_shadow(importName, instance)
+        properties = status.get("properties")
+        if isinstance(properties, dict):
+            instance._properties = properties
         return instance
 
     def status_summary(self):
