@@ -44,6 +44,7 @@ def copy(src):
 def make_map_with_base(doc, baseDir):
     loadTemplate = getattr(doc, "loadTemplate", None)
     _anchorCache = getattr(doc, "_anchorCache", None)
+    lc = getattr(doc, "lc", None)
 
     def factory(*args, **kws):
         if six.PY2 and args and isinstance(args[0], dict):
@@ -55,6 +56,9 @@ def make_map_with_base(doc, baseDir):
             map.loadTemplate = loadTemplate
         if _anchorCache is not None:
             map._anchorCache = _anchorCache
+        if lc is not None:
+            map.lc.line = lc.line
+            map.lc.col = lc.col
         return map
 
     return factory
@@ -262,7 +266,9 @@ def _find_template(doc, key, path, cls, fail):
             if not fail:
                 return None
             else:
-                raise UnfurlError(f"could not find relative path '{('.' * key.relative)}'")
+                raise UnfurlError(
+                    f"could not find relative path '{('.' * key.relative)}'"
+                )
     for index, segment in enumerate(key.pointer):
         if isinstance(template, Sequence):
             # Array indexes should be turned into integers
@@ -437,7 +443,8 @@ def expand_doc(doc, current=None, cls=dict):
             return includes, expanded
         if len(missing) == last:  # no progress
             raise UnfurlError(
-                "missing includes: %s" % [f"{i.key.key}:{i.value or ''}" for i in missing]
+                "missing includes: %s"
+                % [f"{i.key.key}:{i.value or ''}" for i in missing]
             )
         last = len(missing)
         includes = CommentedMap()
