@@ -309,6 +309,11 @@ class Plan:
 
             reason = include(resource)
             if reason:
+                # if the resource doesn't instantiate itself just mark it as absent because otherwise will be in error
+                # (because it will remain with status ok while its dependendents are absent)
+                if resource.template.aggregate_only():
+                    resource._localStatus = Status.absent
+
                 logger.debug("%s instance %s", reason, resource.name)
                 workflow = "undeploy" if reason == Reason.prune else self.workflow
                 yield from self._generate_configurations(resource, reason, workflow)
@@ -436,7 +441,12 @@ class Plan:
         # order by ancestors
         return list(
             order_templates(
-                {t.name: t for t in templates if "virtual" not in t.directives},
+                {
+                    t.name: t
+                    for t in templates
+                    if "virtual" not in t.directives
+                    and interface_requirements_ok(self.tosca, t)
+                },
                 self.filterTemplate and self.filterTemplate.name,
                 self.interface,
             )
@@ -457,7 +467,7 @@ class Plan:
         opts = self.jobOptions
         templates = self._get_templates()
 
-        logger.debug("checking for tasks for templates %s", [t.name for t in templates])
+        logger.verbose("checking for tasks for templates %s", [t.name for t in templates])
         visited = set()
         for template in templates:
             found = False
@@ -740,6 +750,26 @@ def find_explicit_operation_hosts(template, interface):
                 yield operation_host
 
 
+def interface_requirements_ok(spec, template):
+    reqs = template.get_interface_requirements()
+    if reqs:
+        assert isinstance(reqs, list)
+        for req in reqs:
+            if not any(
+                [
+                    rel.is_compatible_type(req)
+                    for rel in spec.topology.default_relationships
+                ]
+            ):
+                logger.debug(
+                    'Skipping template "%s": could not find a connection for interface requirements: %s',
+                    template.name,
+                    reqs,
+                )
+                return False
+    return True
+
+
 def order_templates(templates, filter=None, interface=None):
     # templates is dict of NodeSpecs
     seen = set()
@@ -761,18 +791,21 @@ def order_templates(templates, filter=None, interface=None):
         for ancestor in get_ancestor_templates(source.toscaEntityTemplate):
             spec = templates.get(ancestor.name)
             if spec:
+                if spec is not source:
+                    # ancestor is required by source
+                    spec._isReferencedBy.append(source)
                 if spec in seen:
                     continue
                 seen.add(spec)
-                if spec:
-                    yield spec
+                yield spec
 
 
 def get_ancestor_templates(source):
     # note: opposite direction as NodeSpec.relationships
     for (rel, req, reqDef) in source.relationships:
-        for ancestor in get_ancestor_templates(rel.target):
-            yield ancestor
+        if rel.target is not source:
+            for ancestor in get_ancestor_templates(rel.target):
+                yield ancestor
     yield source
 
 
