@@ -4,6 +4,8 @@ import collections
 import re
 import six
 import shlex
+import sys
+import os
 import os.path
 from .util import (
     lookup_class,
@@ -11,9 +13,10 @@ from .util import (
     find_schema_errors,
     UnfurlError,
     UnfurlTaskError,
+    to_enum
 )
 from .result import serialize_value
-from .support import Defaults
+from .support import Defaults, NodeState
 import logging
 
 logger = logging.getLogger("unfurl")
@@ -38,6 +41,7 @@ class ConfigurationSpec:
             dependencies=None,
             outputs=None,
             interface=None,
+            entry_state=None,
         )
 
     def __init__(
@@ -59,6 +63,7 @@ class ConfigurationSpec:
         dependencies=None,
         outputs=None,
         interface=None,
+        entry_state=None,
     ):
         assert name and className, "missing required arguments"
         self.name = name
@@ -78,6 +83,7 @@ class ConfigurationSpec:
         self.artifact = primary
         self.dependencies = dependencies
         self.interface = interface
+        self.entry_state = to_enum(NodeState, entry_state, NodeState.creating)
 
     def find_invalidate_inputs(self, inputs):
         if not self.inputSchema:
@@ -93,7 +99,8 @@ class ConfigurationSpec:
         return find_schema_errors(expanded, self.preConditions)
 
     def create(self):
-        klass = lookup_class(self.className)
+        className = self.className
+        klass = lookup_class(className)
         if not klass:
             raise UnfurlError(f"Could not load configurator {self.className}")
         else:
@@ -463,12 +470,20 @@ def _render_request(job, parent, req, requests):
     task.target.root.attributeManager = task._attributeManager
 
     error = None
+    error_info = None
     try:
         task.logger.debug("rendering %s %s", task.target.name, task.name)
+        task._rendering = True
+        if task._inputs:
+            # turn strictness off so we can detect dependencies
+            task._inputs.context.strict = False
         task.rendered = task.configurator.render(task)
     except Exception:
         # note: failed rendering may be re-tried later if it has dependencies
-        error = UnfurlTaskError(task, "Configurator render failed", logging.DEBUG)
+        error_info = sys.exc_info()
+        error = UnfurlTaskError(task, "Configurator render failed", False)
+    task._rendering = False
+    task._attributeManager.mark_referenced_templates(task.target.template)
 
     if parent and parent.workflow == "undeploy":
         # when removing an instance don't worry about depending values changing in the future
@@ -476,6 +491,7 @@ def _render_request(job, parent, req, requests):
     else:
         # key => (instance, list<attribute>)
         liveDependencies = task._attributeManager.find_live_dependencies()
+        logger.trace(f"liveDependencies for {task.target}: {liveDependencies}")
         # a future request may change the value of these attributes
         deps = list(_get_deps(parent, req, liveDependencies, requests))
 
@@ -485,7 +501,7 @@ def _render_request(job, parent, req, requests):
             "%s:%s can not render yet, depends on %s",
             task.target.name,
             req.configSpec.operation,
-            str(deps),
+            str(deps), exc_info=error_info
         )
         # rollback changes:
         task._errors = []
@@ -496,6 +512,7 @@ def _render_request(job, parent, req, requests):
     elif error:
         task.fail_work_folders()
         task._inputs = None
+        task.logger.warning("Configurator render failed",  exc_info=error_info)
         task._attributeManager.attributes = {}  # rollback changes
     else:
         task.commit_changes()
@@ -510,6 +527,17 @@ def _add_to_req_list(reqs, parent, request):
         reqs.append(request)
 
 
+def _reevaluate_not_required(not_required, render_requests):
+    # keep rendering if a not_required template was referenced and is now required
+    new_not_required = []
+    for (parent, request) in not_required:
+        if request.target.template.required:
+            render_requests.append( (parent, request) )
+        else:
+            new_not_required.append( (parent, request) )
+    return new_not_required
+
+
 def do_render_requests(job, requests):
     ready, notReady, errors = [], [], []
     flattened_requests = list(
@@ -517,20 +545,31 @@ def do_render_requests(job, requests):
         for (p, r) in get_render_requests(requests)
         if _prepare_request(job, r, errors)
     )
+    not_required = []
     render_requests = collections.deque(flattened_requests)
     while render_requests:
         parent, request = render_requests.popleft()
-        deps, error = _render_request(job, parent, request, flattened_requests)
-        if error:
-            errors.append(error)
-        if deps:
-            # remove if we already added the parent
-            if parent and ready and ready[-1] is parent:
-                ready.pop()
-            _add_to_req_list(notReady, parent, request)
-        elif not parent or not notReady or notReady[-1] is not parent:
-            # don't add if the parent was placed on the notReady list
-            _add_to_req_list(ready, parent, request)
+        # we dont require default templates that aren't referenced
+        # (but skip this check if the job already specified specific instances)
+        required = job.workflow != "deploy" or job.is_filtered() or request.target.template.required
+        if not required:
+            not_required.append( (parent, request) )
+        else:
+            deps, error = _render_request(job, parent, request, flattened_requests)
+            if error:
+                errors.append(error)
+            if deps:
+                # remove parent from ready if added it there
+                if parent and ready and ready[-1] is parent:
+                    ready.pop()
+                _add_to_req_list(notReady, parent, request)
+            elif not parent or not notReady or notReady[-1] is not parent:
+                # don't add if the parent was placed on the notReady list
+                _add_to_req_list(ready, parent, request)
+            not_required = _reevaluate_not_required(not_required, render_requests)
+
+    for (parent, request) in not_required:
+        _add_to_req_list(notReady, parent, request)
     return ready, notReady, errors
 
 
@@ -583,7 +622,7 @@ def find_resources_from_template_name(root, name):
 def find_parent_template(source):
     for rel, req, reqDef in source.relationships:
         # special case "host" so it can be declared without full set of relationship / capability types
-        if rel.type == "tosca.relationships.HostedOn" or "host" in req:
+        if rel.is_derived_from("tosca.relationships.HostedOn") or "host" in req:
             return rel.target
         return None
 
@@ -626,6 +665,18 @@ def create_instance_from_spec(_manifest, target, rname, resourceSpec):
     # note: if resourceSpec[parent] is set it overrides the parent keyword
     return _manifest.create_node_instance(rname, resourceSpec, parent=parent)
 
+def _maybe_mock(iDef, template):
+    if not os.getenv("UNFURL_MOCK_DEPLOY"):
+        return iDef
+    mock = _find_implementation("Mock", iDef.name, template)
+    if mock:
+        return mock
+    # mock operation not found, so patch iDef
+    if not isinstance(iDef.implementation, dict):
+        # it's a string naming an artifact
+        iDef.implementation = dict(primary=iDef.implementation)
+    iDef.implementation['className'] = "unfurl.configurator.MockConfigurator"
+    return iDef
 
 def create_task_request(
     jobOptions,
@@ -642,6 +693,7 @@ def create_task_request(
     interface, sep, action = operation.rpartition(".")
     iDef = _find_implementation(interface, action, resource.template)
     if iDef and iDef.name != "default":
+        iDef = _maybe_mock(iDef, resource.template)
         # merge inputs
         if inputs:
             inputs = dict(iDef.inputs, **inputs)
@@ -674,7 +726,7 @@ def create_task_request(
         )
     else:
         errorMsg = f'unable to find an implementation for operation "{action}" on node "{resource.template.name}"'
-        logger.debug(errorMsg)
+        logger.trace(errorMsg)
         return None
 
     req = TaskRequest(
@@ -730,6 +782,7 @@ def _set_config_spec_args(kw, target, base_dir):
     artifact = kw["primary"]
     className = kw.get("className")
     if not className and not artifact:  # malformed implementation
+        logger.warning("no artifact or className set on operation for %s: %s", target.name, kw)
         return None
     guessing = False
     if not className:
@@ -767,7 +820,7 @@ def _set_config_spec_args(kw, target, base_dir):
 
 def _get_config_spec_args_from_implementation(iDef, inputs, target, operation_host):
     implementation = iDef.implementation
-    kw = dict(inputs=inputs, outputs=iDef.outputs, operation_host=operation_host)
+    kw = dict(inputs=inputs, outputs=iDef.outputs, operation_host=operation_host, entry_state=iDef.entry_state)
     configSpecArgs = ConfigurationSpec.getDefaults()
     artifactTpl = None
     dependencies = None
@@ -783,7 +836,7 @@ def _get_config_spec_args_from_implementation(iDef, inputs, target, operation_ho
             elif name == "dependencies":
                 dependencies = value
             elif name in configSpecArgs:
-                # sets operation_host, environment, timeout
+                # sets operation_host, environment, timeout, className
                 kw[name] = value
     else:
         # "either because it refers to a named artifact specified in the artifacts section of a type or template,
