@@ -3,16 +3,19 @@
 import os
 import os.path
 from pathlib import Path
+import sys
 import git
-from git.repo.fun import is_git_dir
 import logging
 from six.moves.urllib.parse import urlparse
-from .util import UnfurlError, save_to_file, to_text
+from .util import UnfurlError, save_to_file
 import toscaparser.repositories
 from ruamel.yaml.comments import CommentedMap
 
 logger = logging.getLogger("unfurl")
 
+def is_git_worktree(path, gitDir=".git"):
+    # NB: if work tree is a submodule .git will be a file that looks like "gitdir: ./relative/path"
+    return os.path.exists(os.path.join(path, gitDir))
 
 def normalize_git_url(url):
     if url.startswith("git-local://"):  # truncate url after commit digest
@@ -26,6 +29,11 @@ def normalize_git_url(url):
             return "ssh://" + url.replace(":", "/", 1)
     return url
 
+def normalize_git_url_hard(url):
+    parts = urlparse(normalize_git_url(url))
+    # remove scheme, user, and .git
+    return parts.netloc.rpartition("@")[2] + parts.path.rstrip('.git')
+
 
 def is_url_or_git_path(url):
     if "://" in url and not url.startswith("file:"):
@@ -62,7 +70,7 @@ class _ProgressPrinter(git.RemoteProgress):
     def update(self, op_code, cur_count, max_count=None, message=""):
         # we use print instead of logging because we don't want to clutter logs with this message
         if message and logger.getEffectiveLevel() <= logging.INFO:
-            print(f"fetching from {self.gitUrl}, received: {message} ")
+            print(f"fetching from {self.gitUrl}, received: {message} ", file=sys.stderr)
 
 
 class Repo:
@@ -73,7 +81,7 @@ class Repo:
         """
         current = os.path.abspath(rootDir)
         while current and current != os.sep:
-            if is_git_dir(os.path.join(current, gitDir)):
+            if is_git_worktree(current, gitDir):
                 return GitRepo(git.Repo(current))
             current = os.path.dirname(current)
         return None
@@ -88,7 +96,7 @@ class Repo:
 
     @staticmethod
     def update_git_working_dirs(working_dirs, root, dirs, gitDir=".git"):
-        if gitDir in dirs and is_git_dir(os.path.join(root, gitDir)):
+        if gitDir in dirs and is_git_worktree(root, gitDir):
             assert os.path.isdir(root), root
             repo = GitRepo(git.Repo(root))
             key = os.path.abspath(root)
@@ -179,7 +187,7 @@ class Repo:
 
 def commit_secrets(working_dir, yaml):
     vault = yaml and getattr(yaml.representer, "vault", None)
-    if not vault:
+    if not vault or not vault.secrets:
         return []
     saved = []
     for filepath, dotsecrets in find_dirty_secrets(working_dir):
@@ -286,7 +294,11 @@ class RepoView:
                     except Exception as err:
                         logger.warning("could not decrypt %s: %s", filepath, err)
                         continue
-                    with open(str(target), "w") as f:
+                    target_path = str(target)
+                    dir = os.path.dirname(target_path)
+                    if dir and not os.path.isdir(dir):
+                        os.makedirs(dir)
+                    with open(target_path, "w") as f:
                         f.write(contents)
                     os.utime(target, (stinfo.st_atime, stinfo.st_mtime))
                     logger.verbose("decrypted secret file to %s", target)
@@ -298,7 +310,9 @@ class RepoView:
         assert not self.readOnly
         if self.yaml:
             for saved in self.save_secrets():
-                self.repo.repo.git.add(str(saved.relative_to(self.repo.working_dir)))
+                local_path = str(saved.relative_to(self.repo.working_dir))
+                if not self.repo.is_path_excluded(local_path):
+                    self.repo.repo.git.add(local_path)
         if addAll:
             self.add_all()
         self.repo.repo.index.commit(message)
@@ -475,9 +489,9 @@ class GitRepo(Repo):
         status, stdout, stderr = self.run_cmd(["submodule", "add", gitDir])
         success = not status
         if success:
-            logging.debug("added submodule %s: %s %s", gitDir, stdout, stderr)
+            logger.debug("added submodule %s: %s %s", gitDir, stdout, stderr)
         else:
-            logging.error("failed to add submodule %s: %s %s", gitDir, stdout, stderr)
+            logger.error("failed to add submodule %s: %s %s", gitDir, stdout, stderr)
         return success
 
     def get_initial_revision(self):
@@ -501,6 +515,20 @@ class GitRepo(Repo):
         # https://gitpython.readthedocs.io/en/stable/reference.html?highlight=is_dirty#git.repo.base.Repo.is_dirty
         return self.repo.is_dirty(untracked_files=untracked_files, path=path or None)
 
+    def pull(self, remote="origin", revision=None, ff_only=True):
+        if remote in self.repo.remotes:
+            cmd = ["pull", remote, revision or "HEAD"]
+            if ff_only:
+                cmd.append("--ff-only")
+            code, out, err = self.run_cmd(cmd)
+            if code:
+                logger.info("attempt to pull latest from %s into %s failed: %s %s", self.url, self.working_dir, out, err)
+            else:
+                logger.verbose("pull latest from %s into %s: %s %s", self.url, self.working_dir, out, err)
+            return not code
+        else:
+            return False
+
     def clone(self, newPath):
         # note: repo.clone uses bare path, which breaks submodule path resolution
         cloned = git.Repo.clone_from(
