@@ -1,3 +1,4 @@
+
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
 """
@@ -7,6 +8,7 @@ import collections
 from collections.abc import MutableSequence, Mapping
 import copy
 import os
+import sys
 import os.path
 import six
 import re
@@ -16,7 +18,6 @@ from .eval import RefContext, set_eval_func, Ref, map_value
 from .result import (
     Results,
     ResultsMap,
-    ResultsList,
     Result,
     ExternalValue,
     serialize_value,
@@ -32,6 +33,7 @@ from .util import (
     load_module,
     load_class,
     sensitive,
+    filter_env
 )
 from .merge import intersect_dict, merge_dicts
 from unfurl.projectpaths import get_path
@@ -39,6 +41,8 @@ import ansible.template
 from ansible.parsing.dataloader import DataLoader
 from ansible.utils import unsafe_proxy
 from ansible.utils.unsafe_proxy import wrap_var, AnsibleUnsafeText, AnsibleUnsafeBytes
+from jinja2.runtime import ChainableUndefined
+from toscaparser.elements.portspectype import PortSpec
 
 import logging
 
@@ -129,6 +133,12 @@ set_eval_func(
 )
 
 
+set_eval_func(
+    "portspec",
+    lambda arg, ctx: PortSpec.make(map_value(arg, ctx)),
+)
+
+
 class Templar(ansible.template.Templar):
     def template(self, variable, **kw):
         if isinstance(variable, Results):
@@ -204,7 +214,7 @@ def _get_template_test_reg_ex():
 _clean_regex = _get_template_test_reg_ex()
 
 
-def is_template(val, ctx):
+def is_template(val, ctx=None):
     if isinstance(val, (AnsibleUnsafeText, AnsibleUnsafeBytes)):
         # already evaluated in a template, don't evaluate again
         return False
@@ -248,6 +258,15 @@ def _wrap_sequence(v):
 unsafe_proxy._wrap_sequence = _wrap_sequence
 
 
+class UnfurlUndefined(ChainableUndefined):
+     def __getattr__(self, name):
+          if name == '__UNSAFE__':
+              # AnsibleUndefined should never be assumed to be unsafe
+              # This prevents ``hasattr(val, '__UNSAFE__')`` from evaluating to ``True``
+              raise AttributeError(name)
+          # Return original Undefined object to preserve the first failure context
+          return self
+
 def apply_template(value, ctx, overrides=None):
     if not isinstance(value, six.string_types):
         msg = f"Error rendering template: source must be a string, not {type(value)}"
@@ -280,8 +299,10 @@ def apply_template(value, ctx, overrides=None):
     templar.environment.trim_blocks = False
     # templar.environment.lstrip_blocks = False
     fail_on_undefined = ctx.strict
+    if not fail_on_undefined:
+        templar.environment.undefined = UnfurlUndefined
 
-    vars = _VarTrackerDict(__unfurl=ctx)
+    vars = _VarTrackerDict(__unfurl=ctx, __python_executable=sys.executable)
     vars.update(ctx.vars)
     vars.ctx = ctx
 
@@ -481,11 +502,13 @@ def get_env(args, ctx):
         name = args
         default = None
 
-    return env.get(name, default)
+    return env.get(name, map_value(default, ctx))
 
 
 set_eval_func("get_env", get_env, True)
 
+set_eval_func("to_env", lambda args, ctx: filter_env(map_value(args, ctx), addOnly=True))
+
 _toscaKeywordsToExpr = {
     "SELF": ".",
     "SOURCE": ".source",
@@ -877,6 +900,11 @@ class AttributeManager:
         else:
             self.statuses[resource.key][1] = newvalue
 
+    def mark_referenced_templates(self, template):
+        for (resource, attr) in self.attributes.values():
+            if resource.template is not template and template not in resource.template._isReferencedBy:
+                resource.template._isReferencedBy.append(template)
+
     def get_attributes(self, resource):
         if resource.key not in self.attributes:
             # deepcopy() because lazily created ResultMaps and ResultLists will mutate
@@ -950,7 +978,7 @@ class AttributeManager:
     def commit_changes(self):
         changes = {}
         liveDependencies = {}
-        for resource, attributes in self.attributes.values():
+        for resource, attributes in list(self.attributes.values()):
             overrides, specd = attributes._attributes.split()
             # overrides will only contain:
             #  - properties accessed or added while running a task
@@ -960,7 +988,7 @@ class AttributeManager:
             foundSensitive = []
             live = {}
             # items in overrides of type Result have been accessed during this transaction
-            for key, value in overrides.items():
+            for key, value in list(overrides.items()):
                 if not isinstance(value, Result):
                     # hasn't been accessed so keep it as is
                     _attributes[key] = value
@@ -971,6 +999,7 @@ class AttributeManager:
                     # save the Result not savedValue because we need the ExternalValue
                     live[key] = (isLive, savedValue if is_sensitive else value)
                     if not isLive:
+                        resource._properties[key] = savedValue
                         assert not changed  # changed implies isLive
                         continue  # it hasn't changed and it is part of the spec so don't save it as an attribute
 
