@@ -92,9 +92,23 @@ def create_default_topology():
     )
     return ToscaTemplate(yaml_dict_tpl=tpl)
 
+def _patch(node, patchsrc, quote=False, tpl=None):
+    if tpl is None:
+        tpl = node.toscaEntityTemplate.entity_tpl
+    ctx = RefContext(node, dict(template=tpl))
+    ctx.base_dir = getattr(patchsrc, "base_dir", ctx.base_dir)
+    if quote:
+        patch = copy.deepcopy(patchsrc)
+    else:
+        patch = map_value(patchsrc, ctx)
+    logger.trace("patching node %s was %s", node.name, tpl)
+    patched = patch_dict(tpl, patch, True)
+    logger.trace("patched node %s: now %s", node.name, patched)
+    return patched
 
 class ToscaSpec:
     InstallerType = "unfurl.nodes.Installer"
+    topology = None
 
     def evaluate_imports(self, toscaDef):
         if not toscaDef.get("imports"):
@@ -127,6 +141,28 @@ class ToscaSpec:
             return True
         return False
 
+    def enforce_filters(self):
+        patched = False
+        for nodespec in self.nodeTemplates.values():
+            for req in nodespec.requirements.values():
+                for prop, value in req.get_nodefilter_properties():
+                    # annotate the target's properties
+                    target = req.relationship and req.relationship.target
+                    if target and isinstance(value, dict) and 'eval' in value:
+                        value.setdefault('vars', {})['SOURCE'] = dict(eval="::"+nodespec.name)
+                        patch = dict(properties={prop: value})
+                        _patch(target, patch, quote=True)
+                        patched = True
+                for name, value in req.get_nodefilter_requirements():
+                    # annotate the target's requirements
+                    target = req.relationship and req.relationship.target
+                    if target:
+                        matching_target_req = target.requirements.get(name)
+                        _patch(nodespec, value, tpl=matching_target_req.entity_tpl[name])
+                        patched = True
+
+        return patched
+
     def _overlay(self, overlays):
         def _find_matches():
             ExceptionCollector.start()  # clears previous errors
@@ -151,19 +187,7 @@ class ToscaSpec:
                     )
 
         matches = list(_find_matches())
-
-        def _patch(m):
-            node, patchsrc = m
-            tpl = node.toscaEntityTemplate.entity_tpl
-            ctx = RefContext(node, dict(template=tpl))
-            ctx.base_dir = getattr(patchsrc, "base_dir", ctx.base_dir)
-            patch = map_value(patchsrc, ctx)
-            logger.trace("patching node %s was %s", node.name, tpl)
-            patched = patch_dict(tpl, patch, True)
-            logger.trace("patched node %s: now %s", node.name, patched)
-            return patched
-
-        return [_patch(m) for m in matches]
+        return [_patch(*m) for m in matches]
 
     def _parse_template(self, path, inputs, toscaDef, resolver):
         # need to set a path for the import loader
@@ -174,10 +198,14 @@ class ToscaSpec:
             import_resolver=resolver,
             verify=False,  # we display the error messages ourselves so we don't need to verify here
         )
+        ExceptionCollector.collecting = True # don't stop collecting validation errors
+        ExceptionCollector.near = ' while instantiating the spec'
         self.nodeTemplates = {}
         self.installers = {}
         self.relationshipTemplates = {}
         for template in self.template.nodetemplates:
+            if not template.type_definition:
+                continue # invalidate template
             nodeTemplate = NodeSpec(template, self)
             if template.is_derived_from(self.InstallerType):
                 self.installers[template.name] = nodeTemplate
@@ -189,6 +217,11 @@ class ToscaSpec:
                 self.relationshipTemplates[template.name] = relTemplate
         self.load_imported_default_templates()
         self.topology = TopologySpec(self, inputs)
+        substitution_mappings = self.template.topology_template.substitution_mappings
+        if substitution_mappings and substitution_mappings.node:
+            self.substitution_template = self.nodeTemplates.get(substitution_mappings.node)
+        else:
+            self.substitution_template = None
         self.load_workflows()
         self.groups = {
             g.name: GroupSpec(g, self) for g in self.template.topology_template.groups
@@ -197,6 +230,34 @@ class ToscaSpec:
             p.name: PolicySpec(p, self)
             for p in self.template.topology_template.policies
         }
+        ExceptionCollector.collecting = False
+
+    def _patch(self, toscaDef, path):
+        matches = None
+        decorators = self.load_decorators()
+        if decorators:
+            logger.debug("applying decorators %s", decorators)
+            # copy errors before we clear them in _overlay
+            errorsSoFar = ExceptionCollector.exceptions[:]
+            # overlay uses ExceptionCollector
+            matches = self._overlay(decorators)
+            if ExceptionCollector.exceptionsCaught():
+                # abort if overlay caused errors
+                # report previously collected errors too
+                ExceptionCollector.exceptions[:0] = errorsSoFar
+                message = "\n".join(
+                    ExceptionCollector.getExceptionsReport(
+                        full=(get_console_log_level() < logging.INFO)
+                    )
+                )
+                raise UnfurlValidationError(
+                    f"TOSCA validation failed for {path}: \n{message}",
+                    ExceptionCollector.getExceptions(),
+                )
+        modified_imports = self.evaluate_imports(toscaDef)
+        annotated = self.enforce_filters()
+        return matches or modified_imports or annotated
+
 
     def __init__(
         self, toscaDef, spec=None, path=None, resolver=None, skip_validation=False
@@ -224,44 +285,27 @@ class ToscaSpec:
             try:
                 self._parse_template(path, inputs, toscaDef, resolver)
             except:
-                if not ExceptionCollector.exceptionsCaught() or not self.template:
+                if not ExceptionCollector.exceptionsCaught() or not self.template or not self.topology:
                     raise  # unexpected error
 
-            decorators = self.load_decorators()
-            if decorators:
-                logger.debug("applying decorators %s", decorators)
-                # copy errors before we clear them in _overlay
-                errorsSoFar = ExceptionCollector.exceptions[:]
-                self._overlay(decorators)
-                if not skip_validation and ExceptionCollector.exceptionsCaught():
-                    # abort if overlay caused errors
-                    # report previously collected errors too
-                    ExceptionCollector.exceptions[:0] = errorsSoFar
-                    message = "\n".join(
-                        ExceptionCollector.getExceptionsReport(
-                            full=(get_console_log_level() < logging.INFO)
-                        )
-                    )
-                    raise UnfurlValidationError(
-                        f"TOSCA validation failed for {path}: \n{message}",
-                        ExceptionCollector.getExceptions(),
-                    )
-
-            modified_imports = self.evaluate_imports(toscaDef)
-            if decorators or modified_imports:
+            patched = self._patch(toscaDef, path)
+            if patched:
                 # overlay and evaluate_imports modifies tosaDef in-place, try reparsing it
                 self._parse_template(path, inputs, toscaDef, resolver)
 
-            if not skip_validation and ExceptionCollector.exceptionsCaught():
+            if ExceptionCollector.exceptionsCaught():
                 message = "\n".join(
                     ExceptionCollector.getExceptionsReport(
                         full=(get_console_log_level() < logging.INFO)
                     )
                 )
-                raise UnfurlValidationError(
-                    f"TOSCA validation failed for {path}: \n{message}",
-                    ExceptionCollector.getExceptions(),
-                )
+                if skip_validation:
+                    logger.warning("Found TOSCA validation failures: %s", message)
+                else:
+                    raise UnfurlValidationError(
+                        f"TOSCA validation failed for {path}: \n{message}",
+                        ExceptionCollector.getExceptions(),
+                    )
 
     @property
     def base_dir(self):
@@ -281,7 +325,7 @@ class ToscaSpec:
                     return manifest.localEnv.project.projectRoot
         return None
 
-    def add_node_template(self, name, tpl):
+    def add_node_template(self, name, tpl, discovered=True):
         custom_types = None
         if "custom_types" in tpl:
             custom_types = tpl.pop("custom_types")
@@ -292,9 +336,10 @@ class ToscaSpec:
         nodeTemplate = self.template.topology_template.add_template(name, tpl)
         nodeSpec = NodeSpec(nodeTemplate, self)
         self.nodeTemplates[name] = nodeSpec
-        if self.discovered is None:
-            self.discovered = CommentedMap()
-        self.discovered[name] = tpl
+        if discovered:
+            if self.discovered is None:
+                self.discovered = CommentedMap()
+            self.discovered[name] = tpl
         # add custom_types back for serialization later
         if custom_types:
             tpl["custom_types"] = custom_types
@@ -310,7 +355,7 @@ class ToscaSpec:
         return decorators
 
     def load_imported_default_templates(self):
-        for topology in self.template.nested_topologies.values():
+        for name, topology in self.template.nested_topologies.items():
             for nodeTemplate in topology.nodetemplates:
                 if (
                     "default" in nodeTemplate.directives
@@ -550,6 +595,7 @@ class EntitySpec(ResourceRef):
             )
 
         self.type = toscaNodeTemplate.type
+        self._isReferencedBy = [] # this is referenced by another template or via property traversal
         # nodes have both properties and attributes
         # as do capability properties and relationships
         # but only property values are declared
@@ -591,12 +637,8 @@ class EntitySpec(ResourceRef):
         else:
             self.defaultAttributes = {}
 
-    parent = None
-
-    def __reflookup__(self, key):
+    def _resolve(self, key):
         """Make attributes available to expressions"""
-        if key[0] == ".":
-            return self._get_prop(key)
         if key in ["name", "type", "uri", "groups", "policies"]:
             return getattr(self, key)
         raise KeyError(key)
@@ -604,6 +646,11 @@ class EntitySpec(ResourceRef):
     def get_interfaces(self):
         return self.toscaEntityTemplate.interfaces
 
+    def get_interface_requirements(self):
+        return self.toscaEntityTemplate.type_definition.get_interface_requirements(
+            self.toscaEntityTemplate.entity_tpl
+        )
+
     @property
     def groups(self):
         if not self.spec:
@@ -717,6 +764,38 @@ class EntitySpec(ResourceRef):
         else:
             return None
 
+    def aggregate_only(self):
+        "The template is only the sum of its parts."
+        for iDef in self.get_interfaces():
+            if iDef.interfacename == "Standard":
+                return False
+            if iDef.interfacename == "Install" and iDef.name == "discover":
+                return False
+        # no implementations found
+        return True
+
+    @property
+    def required(self):
+        # if this template is required by another template
+        for root in _get_roots(self):
+            if self.spec.substitution_template:
+                if self.spec.substitution_template is root:
+                    # if don't require if a root is the substitution_mappings
+                    return True
+            elif 'default' not in root.directives:
+                # if don't require if this only has defaults templates as a root
+                return True
+        return False
+
+def _get_roots(node, seen=None):
+    # node can reference each other's properties, so we need to handle circular references
+    if seen is None:
+        seen = set()
+    yield node
+    for parent in node._isReferencedBy:
+        if parent.name not in seen:
+            seen.add( node.name )
+            yield from _get_roots(parent, seen)
 
 class NodeSpec(EntitySpec):
     # has attributes: tosca_id, tosca_name, state, (3.4.1 Node States p.61)
@@ -734,9 +813,9 @@ class NodeSpec(EntitySpec):
         self._relationships = []
         self._artifacts = None
 
-    def __reflookup__(self, key):
+    def _resolve(self, key):
         try:
-            return super().__reflookup__(key)
+            return super()._resolve(key)
         except KeyError:
             req = self.get_requirement(key)
             if not req:
@@ -775,13 +854,16 @@ class NodeSpec(EntitySpec):
         if self._requirements is None:
             self._requirements = {}
             nodeTemplate = self.toscaEntityTemplate
-            for (relTpl, req, reqDef) in nodeTemplate.relationships:
+            for (relTpl, req, req_type_def) in nodeTemplate.relationships:
                 name, values = next(iter(req.items()))
-                reqSpec = RequirementSpec(name, req, self)
+                reqSpec = RequirementSpec(name, req, self, req_type_def)
                 if relTpl.target:
                     nodeSpec = self.spec.get_template(relTpl.target.name)
-                    assert nodeSpec
-                    nodeSpec.add_relationship(reqSpec)
+                    if nodeSpec:
+                        nodeSpec.add_relationship(reqSpec)
+                    else:
+                        msg = f'Missing target node "{relTpl.target.name}" for requirement "{name}" on "{self.name}"'
+                        ExceptionCollector.appendException(UnfurlValidationError(msg))
                 self._requirements[name] = reqSpec
         return self._requirements
 
@@ -806,11 +888,13 @@ class NodeSpec(EntitySpec):
         return self._get_relationship_specs()
 
     def _get_relationship_specs(self):
-        if len(self._relationships) != len(self.toscaEntityTemplate.get_relationship_templates()):
+        if len(self._relationships) != len(
+            self.toscaEntityTemplate.get_relationship_templates()
+        ):
             # get_relationship_templates() is a list of RelationshipTemplates that target the node
             rIds = {id(r.toscaEntityTemplate) for r in self._relationships}
             for r in self.toscaEntityTemplate.get_relationship_templates():
-                if id(r) not in rIds:
+                if id(r) not in rIds and r.capability:
                     self._relationships.append(RelationshipSpec(r, self.spec, self))
         return self._relationships
 
@@ -844,14 +928,15 @@ class NodeSpec(EntitySpec):
                 relSpec.toscaEntityTemplate.source.name
                 == reqSpec.parentNode.toscaEntityTemplate.name
             ):
-                assert not reqSpec.relationship or reqSpec.relationship is relSpec
+                assert not reqSpec.relationship or reqSpec.relationship is relSpec, (reqSpec.relationship, relSpec)
                 reqSpec.relationship = relSpec
-                assert not relSpec.requirement or relSpec.requirement is reqSpec
-                relSpec.requirement = reqSpec
+                assert not relSpec.requirement or relSpec.requirement is reqSpec, (relSpec.requirement, reqSpec)
+                if not relSpec.requirement:
+                    relSpec.requirement = reqSpec
                 break
         else:
             msg = f'relationship not found for requirement "{reqSpec.name}" on "{reqSpec.parentNode}" targeting "{self.name}"'
-            raise UnfurlValidationError(msg)
+            ExceptionCollector.appendException(UnfurlValidationError(msg))
 
     @property
     def abstract(self):
@@ -909,9 +994,9 @@ class RelationshipSpec(EntitySpec):
     def target(self):
         return self.capability.parentNode if self.capability else None
 
-    def __reflookup__(self, key):
+    def _resolve(self, key):
         try:
-            return super().__reflookup__(key)
+            return super()._resolve(key)
         except KeyError:
             if self.capability:
                 if self.capability.parentNode.is_compatible_target(key):
@@ -950,17 +1035,21 @@ class RequirementSpec:
     """
 
     # XXX need __eq__ since this doesn't derive from EntitySpec
-    def __init__(self, name, req, parent):
+    def __init__(self, name, req, parent, type_tpl):
         self.source = self.parentNode = parent  # NodeSpec
         self.spec = parent.spec
         self.name = name
         self.entity_tpl = req
         self.relationship = None
+        self.type_tpl = type_tpl
         # entity_tpl may specify:
         # capability (definition name or type name), node (template name or type name), and node_filter,
         # relationship (template name or type name or inline relationship template)
         # occurrences
 
+    def __repr__(self):
+        return f"{self.__class__.__name__}('{self.name}')"
+
     @property
     def artifacts(self):
         return self.parentNode.artifacts
@@ -971,6 +1060,22 @@ class RequirementSpec:
     def get_interfaces(self):
         return self.relationship.get_interfaces() if self.relationship else []
 
+    def get_nodefilter_properties(self):
+        # XXX should merge type_tpl with entity_tpl
+        return get_nodefilters(self.type_tpl, 'properties')
+
+    def get_nodefilter_requirements(self):
+        # XXX should merge type_tpl with entity_tpl
+        return get_nodefilters(self.type_tpl, "requirements")
+
+def get_nodefilters(entity_tpl, key):
+    if not isinstance(entity_tpl, dict):
+        return
+    nodefilter = entity_tpl.get('node_filter')
+    if nodefilter and key in nodefilter:
+        for filter in nodefilter[key]:
+            name, value = next(iter(filter.items()))
+            yield name, value
 
 class CapabilitySpec(EntitySpec):
     def __init__(self, parent=None, capability=None):
@@ -1050,6 +1155,8 @@ class TopologySpec(EntitySpec):
         self.propertyDefs = {}
         self.attributeDefs = {}
         self.capabilities = []
+        self._defaultRelationships = None
+        self._isReferencedBy = []
 
     def get_interfaces(self):
         # doesn't have any interfaces
@@ -1067,14 +1174,24 @@ class TopologySpec(EntitySpec):
     def primary_provider(self):
         return self.spec.relationshipTemplates.get("primary_provider")
 
+    @property
+    def default_relationships(self):
+        if self._defaultRelationships is None:
+            self._defaultRelationships = [
+                relSpec
+                for relSpec in self.spec.relationshipTemplates.values()
+                if relSpec.toscaEntityTemplate.default_for
+            ]
+        return self._defaultRelationships
+
     @property
     def base_dir(self):
         return self.spec.base_dir
 
-    def __reflookup__(self, key):
+    def _resolve(self, key):
         """Make attributes available to expressions"""
         try:
-            return super().__reflookup__(key)
+            return super()._resolve(key)
         except KeyError:
             nodeTemplates = self.spec.nodeTemplates
             nodeSpec = nodeTemplates.get(key)
