@@ -4,8 +4,8 @@ import os.path
 import sys
 import codecs
 import json
-import functools
-from typing import Union
+import os
+from typing import Union, Tuple, List
 import six
 from six.moves import urllib
 
@@ -103,6 +103,8 @@ def represent_sensitive_bytes(dumper, data):
 def _construct_vault(constructor, node, tag):
     value = constructor.construct_scalar(node)
     if not constructor.vault.secrets:
+        if os.getenv("UNFURL_VAULT_SKIP_DECRYPT"):
+            return sensitive_str.redacted_str
         raise ConstructorError(
             context=None,
             context_mark=None,
@@ -115,18 +117,24 @@ def _construct_vault(constructor, node, tag):
 
 def construct_vault(constructor, node):
     value = _construct_vault(constructor, node, "!vault")
+    if value is sensitive_str.redacted_str:
+        return value
     cleartext = to_text(constructor.vault.decrypt(value))
     return sensitive_str(cleartext)
 
 
 def construct_vaultjson(constructor, node):
     value = _construct_vault(constructor, node, "!vault-json")
+    if value is sensitive_str.redacted_str:
+        return value
     cleartext = to_text(constructor.vault.decrypt(value))
     return wrap_sensitive_value(json.loads(cleartext))
 
 
 def construct_vaultbinary(constructor, node):
     value = _construct_vault(constructor, node, "!vault-binary")
+    if value is sensitive_str.redacted_str:
+        return value
     cleartext = constructor.vault.decrypt(value)
     return sensitive_bytes(cleartext)
 
@@ -172,7 +180,7 @@ def make_vault_lib(passwordBytes, vaultId="default"):
     return None
 
 
-def make_vault_lib_ex(secrets: [str, Union[str, bytes]]):
+def make_vault_lib_ex(secrets: List[Tuple[str, Union[str, bytes]]]):
     vault_secrets = []
     for vaultId, passwordBytes in secrets:
         if isinstance(passwordBytes, six.string_types):
@@ -266,7 +274,7 @@ class ImportResolver(toscaparser.imports.ImportResolver):
 
     def load_yaml(self, importLoader, path, isFile=True, fragment=None):
         try:
-            logger.debug(
+            logger.trace(
                 "attempting to load YAML %s: %s", "file" if isFile else "url", path
             )
             originalPath = path
@@ -318,7 +326,7 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                         doc.includes = includes
                     doc.path = path
 
-            if fragment:
+            if fragment and doc:
                 return _refResolver.resolve_fragment(doc, fragment)
             else:
                 return doc
@@ -404,10 +412,10 @@ class YamlConfig:
         path = os.path.abspath(os.path.join(baseDir or self.get_base_dir(), path))
         if warnWhenNotFound and not os.path.isfile(path):
             return path, None
-        logger.debug("attempting to load YAML file: %s", path)
+        logger.trace("attempting to load YAML file: %s", path)
         with open(path, "r") as f:
             config = self.yaml.load(f)
-        if fragment:
+        if fragment and config:
             return path, _refResolver.resolve_fragment(config, fragment)
         return path, config
 
@@ -523,6 +531,9 @@ class YamlConfig:
                     f"`file` key missing from document include: {templatePath}"
                 )
             key = templatePath["file"]
+            when = templatePath.get("when")
+            if when and not os.getenv(when):
+                return value, None, ''
         else:
             value = None
             key = templatePath
