@@ -314,16 +314,35 @@ class YamlManifest(ReadOnlyManifest):
 
         spec = manifest.get("spec", {})
         more_spec = self._load_context(self.context, localEnv)
-        # if we are exporting (thus skip_validation is True)
-        if skip_validation and spec.get("resource_templates"):
-            node_templates = more_spec["topology_template"]["node_templates"]
-            self._load_resource_templates(spec["resource_templates"], node_templates)
+        deployment_blueprint = self.context.get("deployment_blueprint")
+        deployment_blueprints = (
+            manifest.get("spec", {}).get("deployment_blueprints") or {}
+        )
+        if deployment_blueprint:
+            if deployment_blueprint not in deployment_blueprints:
+                raise UnfurlError(
+                    f"Can not find requested deployment blueprint: '{deployment_blueprint}' is missing from the ensemble."
+                )
+            logger.info('Using deployment blueprint "%s"', deployment_blueprint)
+            resource_templates = deployment_blueprints[deployment_blueprint].get("resource_templates")
+            if resource_templates:
+                node_templates = more_spec["topology_template"]["node_templates"]
+                self._load_resource_templates(resource_templates, node_templates, False)
+        elif deployment_blueprints:
+            logger.warning("This ensemble contains deployment blueprints but none were specified for use.")
         if self.context.get("instances"):
             # add context instances to spec instances
-            self._load_resource_templates(self.context["instances"], spec.setdefault('instances', {}))
+            self._load_resource_templates(
+                self.context["instances"], spec.setdefault("instances", {}), True
+            )
 
         self._set_spec(spec, more_spec, skip_validation)
         assert self.tosca
+        if self.localEnv:
+            msg = f'Loading ensemble in environment "{self.localEnv.manifest_context_name}"'
+            if self.tosca.topology and self.tosca.topology.primary_provider:
+                msg += f' with a primary_provider of type "{self.tosca.topology.primary_provider.type}"'
+            logger.info(msg)
 
         status = manifest.get("status", {})
         self.changeLogPath = manifest.get("jobsLog")
@@ -437,14 +456,15 @@ class YamlManifest(ReadOnlyManifest):
             self.create_node_instance(key, val, root)
         return root
 
-    def _load_resource_templates(self, templates, node_templates):
+    def _load_resource_templates(self, templates, node_templates, virtual):
         # "resource_templates" are node templates that aren't included in the topology_template
         # but are referenced by the deployment blueprints
         # or are instances that are part of the environment
         # XXX these might not be node_templates, need to check type
         for name, tpl in templates.items():
             # hacky way to exclude being part of the deployment plan and the manifest's status
-            tpl.setdefault("directives", []).append("virtual")
+            if virtual:
+                tpl.setdefault("directives", []).append("virtual")
             node_templates[name] = tpl
 
     def _load_context(self, context, localEnv):
@@ -472,7 +492,7 @@ class YamlManifest(ReadOnlyManifest):
         location = value.get("manifest")
         if not location:
             raise UnfurlError(
-                f"Can not import external ensemble '{(name)}': no manifest specified"
+                f"Can not import external ensemble '{name}': no manifest specified"
             )
 
         if "project" in location:
@@ -594,6 +614,9 @@ class YamlManifest(ReadOnlyManifest):
         # particularly, because these will get loaded in later runs and mask any spec properties with the same name
         if resource._attributes:
             status["attributes"] = resource._attributes
+        # save computed values for properties as they were observed
+        if resource._properties:
+            status["properties"] = resource._properties
         if resource.shadow:
             # name will be the same as the import name
             status["imported"] = resource.name
@@ -629,7 +652,8 @@ class YamlManifest(ReadOnlyManifest):
 
     def save_resource(self, resource, discovered):
         # XXX checkstatus break unit tests so skip mostly
-        checkstatus = resource.template.type == "unfurl.nodes.LocalRepository"
+        checkstatus = (resource.template.type == "unfurl.nodes.LocalRepository"
+                        or "default" in resource.template.directives)
         ret = self._save_entity_if_instantiated(resource, checkstatus)
         if not ret:
             return ret
